"""This module defines the StateSpace class for representing a linear
continuous time-invariant system as a state-space model.

Copyright 2021--2023 Michael Hayes, UCECE

"""

from .sexpr import LaplaceDomainExpression
from .cache import cached_property
from .matrix import Matrix
from .smatrix import LaplaceDomainMatrix
from .tmatrix import TimeDomainMatrix
from .nmatrix import DiscreteTimeDomainMatrix
from .statespacebase import StateSpaceBase
from .texpr import texpr
from .equation import Equation
from .sym import ssym, tsym
from sympy import diag, eye, MatMul, MatAdd, Derivative


# Note, need to be careful with simplification since a pole can
# cancel a zero.


class StateSpace(StateSpaceBase):
    """Continuous-time linear time-invariant state space model."""

    @property
    def u(self):
        """Input vector."""
        if self._u is None:
            self._u = TimeDomainMatrix([texpr('u_%d(t)' % i) for i in
                                        range(self.Nu)])
        return self._u

    @property
    def x(self):
        """State variable vector."""
        if self._x is None:
            self._x = TimeDomainMatrix([texpr('x_%d(t)' % i) for i in
                                        range(self.Nx)])
        return self._x

    @cached_property
    def dotx(self):
        """Time derivative of state variable vector."""

        return TimeDomainMatrix([Derivative(x1, tsym) for x1 in self.x.sympy])

    @property
    def x0(self):
        """State variable initial value vector."""
        if self._x0 is None:
            self._x0 = TimeDomainMatrix([texpr('x_0_%d(t)' % i) for i in
                                         range(self.Nx)])
        return self._x0

    @property
    def y(self):
        """Output vector."""
        if self._y is None:
            self._y = TimeDomainMatrix([texpr('y_%d(t)' % i) for i in
                                        range(self.Ny)])
        return self._y

    @property
    def U(self):
        """Laplace transform of input vector."""
        return LaplaceDomainMatrix(self.u.LT())

    @property
    def X(self):
        """Laplace transform of state-variable vector."""
        return LaplaceDomainMatrix(self.x.LT())

    @property
    def Y(self):
        """Laplace transform of output vector."""
        return LaplaceDomainMatrix(self.y.LT())

    @cached_property
    def H(self):
        """X(s) / U(s)"""
        return LaplaceDomainMatrix(self.Phi * self._B)

    @property
    def h(self):
        """ILT{X(s) / U(s)}"""
        return TimeDomainMatrix(self.H.ILT(causal=True))

    @cached_property
    def G(self):
        """System transfer functions.
        For a SISO system, use G[0].
        """
        return LaplaceDomainMatrix(self._C * self.H + self._D)

    def state_equations(self):
        """System of first-order differential state equations:

        dotx(t) = A x(t) + B u(t)

        where x is the state vector and u is the input vector.
        """

        if len(self.u) == 0 and len(self.x) == 0:
            return Equation(None, None)

        if len(self.u) == 0:
            return Equation(self.dotx.sympy,
                            MatMul(self._A.sympy, self.x.sympy))

        if len(self.x) == 0:
            return Equation(self.dotx.sympy,
                            MatMul(self._B.sympy, self.u.sympy))

        return Equation(self.dotx.sympy,
                        MatAdd(MatMul(self._A.sympy, self.x.sympy),
                               MatMul(self._B.sympy, self.u.sympy)))

    def output_equations(self):
        """System of output equations:

        y(t) = C x(t) + D u(t)

        where y is the output vector, x is the state vector and u is
        the input vector.

        """

        if len(self.u) == 0:
            return Equation(self.y,
                            MatMul(self._C.sympy, self.x.sympy))

        if len(self.x) == 0:
            return Equation(self.y,
                            MatMul(self._D.sympy, self.u.sympy))

        return Equation(self.y,
                        MatAdd(MatMul(self._C.sympy, self.x.sympy),
                               MatMul(self._D.sympy, self.u.sympy)))

    @property
    def g(self):
        """System impulse responses."""
        return TimeDomainMatrix(self.G.ILT(causal=True))

    @cached_property
    def Phi(self):
        """s-domain state transition matrix."""

        M = LaplaceDomainMatrix(eye(self.Nx) * ssym - self._A.sympy)
        return LaplaceDomainMatrix(M.inv())

    @cached_property
    def phi(self):
        """State transition matrix."""
        return TimeDomainMatrix(self.Phi.ILT(causal=True))

    def characteristic_polynomial(self):
        """Characteristic polynomial (aka system polynomial).

        `lambda(s) = |s * I - A|`"""

        M = LaplaceDomainMatrix(eye(self.Nx) * ssym - self._A.sympy)
        return LaplaceDomainExpression(M.det()).simplify()

    @cached_property
    def P(self):
        """Characteristic polynomial (aka system polynomial).

        `lambda(s) = |s * I - A|`"""

        return self.characteristic_polynomial()

    @cached_property
    def Lambda(self):
        """Diagonal matrix of eigenvalues."""

        # Perhaps faster to use diagonalize
        # E, L = self.A.diagonalize()
        # return L

        e = self.eigenvalues
        return LaplaceDomainMatrix(diag(*e))

    @cached_property
    def M(self):
        """Modal matrix (eigenvectors of A)."""

        E, L = self._A.sympy.diagonalize()

        return LaplaceDomainMatrix(E)

    @cached_property
    def controllability_gramian(self):
        """Controllability gramian matrix."""

        from scipy import linalg

        B = self.B.evaluate()
        Q = -B @ B.T

        # Find Wc given A @ Wc + Wc @ A.T = Q
        # Wc > o if (A, B) controllable
        Wc = linalg.solve_continuous_lyapunov(self.A.evaluate(), Q)

        # Wc should be symmetric positive semi-definite
        Wc = (Wc + Wc.T) / 2

        return Matrix(Wc)

    @property
    def Wc(self):
        """Controllability gramian matrix."""
        return self.controllability_gramian

    @property
    def reachability_gramian(self):
        """Reachability gramian matrix.  This is equivalent to the
        controllability gramian matrix for a linear time independent
        system."""

        return self.controllability_gramian

    @property
    def Wr(self):
        """Reachability gramian matrix."""
        return self.reachability_gramian

    @cached_property
    def observability_gramian(self):
        """Observability gramian matrix."""

        from scipy import linalg

        C = self.C.evaluate()
        Q = -C.T @ C

        # Find Wo given A.T @ Wo + Wo @ A = Q
        # Wo > o if (C, A) observable
        Wo = linalg.solve_continuous_lyapunov(self.A.evaluate().T, Q)

        # Wo should be symmetric positive semi-definite
        Wo = (Wo + Wo.T) / 2

        return Matrix(Wo)

    @property
    def Wo(self):
        """Observability gramian matrix."""
        return self.observability_gramian

    def generalized_bilinear_transform(self, alpha=0.5):

        from .sym import dt
        from .symbols import n
        from .dtstatespace import DTStateSpace

        if alpha < 0 or alpha > 1:
            raise ValueError("alpha must be between 0 and 1 inclusive")

        I = eye(self.Nx)
        M = I - alpha * dt * self.A
        Minv = M.inv()

        Ad = Minv * (I + (1 - alpha) * dt * self.A)
        Bd = Minv * dt * self.B
        Cd = (Minv.T * self.C.T).T
        Dd = self.D + alpha * self.C * Bd

        ud = self._u.discretize(drop_dt=True)
        yd = self._y.discretize(drop_dt=True)
        xd = self._x.discretize(drop_dt=True)
        x0d = self._x0.discretize(drop_dt=True)

        return DTStateSpace(Ad, Bd, Cd, Dd,
                            DiscreteTimeDomainMatrix(ud),
                            DiscreteTimeDomainMatrix(yd),
                            DiscreteTimeDomainMatrix(xd),
                            DiscreteTimeDomainMatrix(x0d))

    def discretize(self, method='bilinear', alpha=0.5):
        """Convert to a discrete-time state space approximation.

        The default method is 'bilinear'.  Other methods are
        'forward_euler', 'backward_euler', and 'gbf'.
        The latter has a parameter `alpha`."""

        if method == 'gbf':
            return self.generalized_bilinear_transform(alpha)
        elif method in ('bilinear', 'tustin'):
            return self.generalized_bilinear_transform(0.5)
        elif method in ('euler', 'forward_diff', 'forward_euler'):
            return self.generalized_bilinear_transform(0)
        elif method in ('backward_diff', 'backward_euler'):
            return self.generalized_bilinear_transform(1)
        else:
            raise ValueError('Unsupported method %s' % method)

    @classmethod
    def from_circuit(cls, cct, node_voltages=None, branch_currents=None):

        from .statespacemaker import StateSpaceMaker

        return StateSpaceMaker.from_circuit(cct, node_voltages, branch_currents)



"""This module provides the NetlistMixin class.  This is used for
Netlist and SubNetlist.

Copyright 2020--2023 Michael Hayes, UCECE

"""

from .analysis import Analysis
from .attrdict import AttrDict
from .components import Components
from .current import Iname
from .deprecation import LcapyDeprecationWarning
from .equipotentialnodes import EquipotentialNodes
from .expr import expr
from .mnacpts import Cpt
from .netfile import NetfileMixin
from .netlistnamespace import NetlistNamespace
from .node import Node
from .nodes import Nodes
from .state import state
from .symbols import j, s, omega
from .voltage import Vname
from . import mnacpts
from collections import OrderedDict
from warnings import warn


class NetlistMixin(object):

    def __init__(self, filename=None, context=None, allow_anon=False):

        self._elements = OrderedDict()
        self.namespaces = {}
        self.nodes = Nodes()
        if context is None:
            context = state.new_context()

        self.context = context
        self.allow_anon = allow_anon
        self._init_parser(mnacpts, allow_anon=allow_anon)

        if filename is not None:
            self.netfile_add(filename)

    def __getitem__(self, name):
        """Return element or node by name."""

        # If name is an integer, convert to a string.
        if isinstance(name, int):
            name = '%d' % name

        if name in self.nodes:
            return self.nodes[name]

        if name in self._elements:
            return self._elements[name]

        # Try first anonymous name.
        if name + 'anon1' in self._elements:
            return self._elements[name + 'anon1']

        if name in self.namespaces:
            return self.namespaces[name]

        raise AttributeError('Unknown element or node name %s' % name)

    def __getattr__(self, attr):
        """Return element or node by name.  This gets called if there is no
        explicit attribute attr for this instance.  This is primarily
        for accessing elements and non-numerical node names.  It also
        gets called if the called attr throws an AttributeError
        exception.  The annoying thing is that hasattr uses getattr
        and checks for an exception."""

        return self.__getitem__(attr)

    def __repr__(self):

        return self.netlist()

    @property
    def analysis(self):

        if not hasattr(self, '_analysis'):
            self._analysis = self.analyse()

        return self._analysis

    @property
    def branch_list(self):
        """Determine list of names of branch elements, e.g.,
        ['C1', 'V1', 'R1', 'R2']."""

        if hasattr(self, '_branch_list'):
            return self._branch_list

        self._branch_list = []
        for key, elt in self.elements.items():
            if elt.type not in ('W', 'O', 'P', 'K', 'XX'):
                self._branch_list.append(elt.name)
        return self._branch_list

    @property
    def capacitors(self):
        """Return list of capacitors."""

        return self.components.capacitors

    @property
    def cg(self):
        """Generate circuit graph for this netlist.  This is cached."""

        return self.circuit_graph()

    @property
    def components(self):

        if not hasattr(self, '_components'):
            self._components = Components(self)

        return self._components

    @property
    def control_sources(self):
        """Return list of voltage sources required to specify control
        current for CCVS and CCCS components."""
        return self.analysis.control_sources

    @property
    def current_sources(self):
        """Return list of current_sources."""

        return self.components.current_sources

    @property
    def dependent_sources(self):
        """Return list of dependent sources."""

        return self.analysis.dependent_sources

    @property
    def elements(self):

        if hasattr(self, '_add_elements'):
            if self._elements == {}:
                self._add_elements()

        return self._elements

    @property
    def equipotential_nodes(self):
        """Determine nodes connected by wires that are of the same potential.
        This returns a dictionary keyed by the unique node names with
        values being lists of nodes of the same potential.

        Primary nodes (without underscore) are chosen for the key
        if possible."""

        enodes = EquipotentialNodes()
        enodes.add(self.nodes.keys())

        # Then augment with nodes connected by wires.
        for m, elt in enumerate(self.elements.values()):
            if elt.nosim:
                continue
            if elt.type == 'W':
                enodes.add_wire(*elt.node_names)
            elif elt.type.startswith('TL'):
                enodes.add_wire(elt.node_names[1], elt.node_names[3])
            elif elt.type.startswith('TP'):
                enodes.add_wire(elt.node_names[1], elt.node_names[3])
                warn("Assuming V2' = V1' for %s" % elt.name)
            else:
                for connections in elt.equipotential_nodes:
                    enodes.add_wires([elt.name + '.' + n for n in connections])

        # Alter keys to avoid underscore and to ensure that have a '0'
        # key if possible.
        enodes2 = {}
        for key, nodes in enodes.items():
            # Sort so that lowest primary nodes come first
            nodes = sorted(nodes, key=lambda item: ('_' in item, item))
            if '0' in nodes:
                newkey = '0'
            else:
                newkey = nodes[0]
            enodes2[newkey] = nodes

        return enodes2

    @property
    def has_dc(self):
        """Return True if any independent source has a DC component."""
        return self.analysis.has_dc

    @property
    def has_ac(self):
        """Return True if any independent source has an AC component."""
        return self.analysis.has_ac

    @property
    def has_s_transient(self):
        """Return True if any independent source has a transient component defined in Laplace-domain."""
        return self.analysis.has_s

    @property
    def has_transient(self):
        """Return True if any independent source has a transient component."""
        return self.analysis.has_transient

    @property
    def is_causal(self):
        """Return True if all independent sources are causal and not an
        initial value problem (unless all the initial values are zero)."""
        return self.analysis.causal

    @property
    def is_connected(self):
        """Return True if all components are connected."""

        return self.cg.is_connected

    @property
    def is_dc(self):
        """Return True if all independent sources are DC and not an
        initial value problem.  The initial value problem may collapse
        to a DC problem but we cannot prove this yet."""
        return self.analysis.dc

    @property
    def ics(self):
        """Return list of components with initial conditions."""

        return self.analysis.ics

    @property
    def independent_sources(self):
        """Return list of independent sources (this does not include
        implicit sources due to initial conditions)."""

        return self.analysis.independent_sources

    def independent_source_groups(self, transform=False):
        """Return list of source groups.  Each group is a list of
        sourcenames that can be analysed at the same time.  Noise
        sources have separate groups since they are assumed to be
        uncorrelated.  The source groups are keyed by the names of
        superposition decomposition keys: 'dc', 's', 't', 'n*', and omega,
        where omega is an expression for the angular frequency of a phasor,
        and `n*` is a noise identifier.

        If transform is False, the returned keys are 'dc', 't', 's', and 'n*'.

        If transform is True, the returned keys are 'dc', 's', omega,
        and 'n*'.  Note, there is no time-domain component.  Note,
        after transformation, a source can appear in multiple groups.
        For example, if a voltage source V1 has a value 10 + 5 *
        cos(omega * t), V1 will be added to the dc and omega groups.

        """

        groups = {}
        for eltname, elt in self.elements.items():
            if not elt.is_independent_source:
                continue
            cpt = elt.cpt
            if cpt.is_voltage_source:
                Voc = cpt.Voc
                if transform:
                    Voc = Voc.decompose()
                cpt_kinds = Voc.keys()
            else:
                Isc = cpt.Isc
                if transform:
                    Isc = Isc.decompose()
                cpt_kinds = Isc.keys()

            for cpt_kind in cpt_kinds:
                if cpt_kind not in groups:
                    groups[cpt_kind] = []
                groups[cpt_kind].append(eltname)

        return groups

    @property
    def inductors(self):
        """Return list of inductors."""

        return self.components.inductors

    @property
    def is_ac(self):
        """Return True if all independent sources are AC and not an
        initial value problem."""
        return self.analysis.ac

    @property
    def is_IVP(self):
        """Return True for an initial value problem.  This is True if any
        component (L, C) that allows initial conditions has them explicitly
        defined.

        """
        return self.analysis.ivp

    @property
    def is_ivp(self):
        LcapyDeprecationWarning(
            feature="is_ivp",
            useinstead="is_IVP",
            deprecated_since_version="1.7"
        ).warn()
        return self.is_IVP

    @property
    def is_passive(self):
        """Return True for a passive network, i.e., there are no sources."""

        return self.sources == {}

    @property
    def is_superposition(self):
        """Return True if netlist needs to be solved using multiple approaches,
        e.g., ac and dc"""

        return ((self.has_ac and self.has_dc) or
                (self.has_ac and self.has_transient) or
                (self.has_dc and self.has_transient))

    @property
    def is_switching(self):
        """Return True for a switching circuit."""
        return self.analysis.switching

    @property
    def is_time_domain(self):
        """Return True if can analyse in time domain."""
        return self.analysis.time_domain

    @property
    def missing_ic(self):
        """Return components that allow initial conditions but do not have
        them explicitly defined."""

        return dict((key, cpt) for key, cpt in self.elements.items() if cpt.has_ic is False)

    @property
    def mutual_couplings(self):
        """Return list of mutual couplingss."""

        return self.analysis.mutual_couplings

    @property
    def node_list(self):
        """Determine list of sorted unique node names, e.g.,
        ['0', '1', '2']."""

        if hasattr(self, '_node_list'):
            return self._node_list

        # Extract unique nodes.
        node_list = list(self.equipotential_nodes.keys())
        node_list = sorted(node_list)
        # Ensure node '0' is first in the list.
        if '0' in node_list:
            node_list.insert(0, node_list.pop(node_list.index('0')))

        self._node_list = node_list
        return node_list

    @property
    def node_map(self):
        """Create dictionary mapping node names to the unique
        equipotential node names."""

        if hasattr(self, '_node_map'):
            return self._node_map

        enodes = self.equipotential_nodes

        # Create inverted dictionary that maps the node names
        # to the equipotential node names.
        node_map = {}
        for key, nodes in enodes.items():
            for node in nodes:
                node_map[node] = key

        self._node_map = node_map
        return node_map

    @property
    def reactances(self):
        """Return list of reactances."""

        return self.analysis.reactances

    @property
    def sch(self):
        """Generate schematic of subnetlist."""

        from .schematic import Schematic

        if hasattr(self, '_sch'):
            return self._sch

        sch = Schematic(allow_anon=self.allow_anon)

        netlist = self.netlist()
        for net in netlist.split('\n'):
            sch.add(net)

        sch.subnetlists = self.subnetlists

        # Don't cache schematic; need to remove splitting of implicit nodes
        # from the draw method otherwise a.draw(); a.draw() fails.
        #
        # self._sch = sch
        return sch

    @property
    def sources(self):
        """Return list of all sources (this does not include
        implicit sources due to initial conditions)."""

        return self.dependent_sources + self.independent_sources

    @property
    def transformers(self):
        """Return list of transformers."""

        return self.components.transformers

    @property
    def voltage_sources(self):
        """Return list of voltage_sources."""

        return self.components.voltage_sources

    @property
    def zeroic(self):
        """Return True if the initial conditions for all components are zero."""
        return self.analysis.zeroic

    def _check_nodes(self, *nodes):

        str_nodes = []
        for node in nodes:
            if isinstance(node, int):
                node = '%s' % node
            if node not in self.nodes:
                raise ValueError('Unknown node %s' % node)
            str_nodes.append(node)
        return str_nodes

    def _dummy_node_name(self):
        """Create a dummy node name."""

        return '_' + self._make_anon_node_name()

    def _find_combine_subsets(self, aset):
        """Return dict of subsets of component names where each subset has the
        same component type."""

        aset = aset.copy()

        subsets = {}
        while aset != set():
            name = aset.pop()
            cpt = self._elements[name]
            aset.add(name)

            subset = set()
            for name1 in aset:
                cpt1 = self._elements[name1]
                if cpt.type == cpt1.type:
                    subset.add(name1)
            aset -= subset
            if len(subset) > 1:
                subsets[cpt.type] = subset
        return subsets

    def _in_parallel_all(self):

        names = self._potential_combine_names()
        lists = []
        while names != []:
            name = names[0]
            parallel = self._in_parallel_names(name)
            if len(parallel) > 1:
                lists.append(parallel)
            for name in parallel:
                try:
                    names.remove(name)
                except:
                    pass
        return lists

    def _in_series_all(self):

        names = self._potential_combine_names()
        lists = []
        while names != []:
            name = names[0]
            series = self._in_series_names(name)
            if len(series) > 1:
                lists.append(series)
            for name in series:
                try:
                    names.remove(name)
                except:
                    pass
        return lists

    def _in_parallel_names(self, cpt=None):

        if isinstance(cpt, Cpt):
            cpt = cpt.name

        parallel_set = self.cg.in_parallel(cpt)
        return parallel_set

    def _in_series_names(self, cpt=None):

        if isinstance(cpt, Cpt):
            cpt = cpt.name

        series_set = self.cg.in_series(cpt)
        return series_set

    def _initialize_from_dict(self, values):

        for key, value in values.items():
            if key not in self.reactances:
                warn('Cannot set initial value for %s since not a reactance' % key)
            elif key not in self._elements:
                raise ValueError('Unknown compoment %s' % key)

        new = self._new()

        for cpt in self._elements.values():
            ic = 0
            if cpt.name in values:
                ic = values[cpt.name]
                try:
                    ic = ic.remove_condition()
                except:
                    pass

            net = cpt._initialize(ic)
            new._add(net)
        return new

    def _initialize_from_circuit(self, cct, T=None):

        if T is None:
            raise ValueError('Time T not specified')

        new = self._new()

        for cpt in self._elements.values():
            ic = 0
            if cpt.name in cct.reactances:
                if cpt.type == 'C':
                    ic = cct[cpt.name].v.remove_condition().subs(T)
                else:
                    ic = cct[cpt.name].i.remove_condition().subs(T)

            net = cpt._initialize(ic)
            new._add(net)
        return new

    def _invalidate(self):

        for attr in ('_sch', '_sub', '_Vdict', '_Idict', '_analysis',
                     '_node_map', '_ss', '_node_list', '_branch_list', '_cg',
                     '_mna', '_la', '_na'):
            try:
                delattr(self, attr)
            except:
                pass

    def _kill(self, sourcenames):

        new = self._new()

        for cpt in self._elements.values():
            if cpt.name in sourcenames:
                if cpt.name in self.control_sources:
                    net = cpt._zero()
                else:
                    net = cpt._kill()
            else:
                net = cpt._copy()
            new._add(net)
        return new

    def _namespace_add(self, namespace):

        namespace = namespace.strip('.')
        if namespace == '':
            return

        parts = namespace.split('.')
        namespaces = self.namespaces
        namespace = ''
        for part in parts:
            if part not in namespaces:
                namespace += part
                namespaces[part] = NetlistNamespace(namespace, self)
                namespace += '.'
                namespaces = namespaces[part].namespaces

    def _noisy(self, resistornames, T='T'):

        new = self._new()

        for cpt in self._elements.values():
            if cpt.name in resistornames:
                net = cpt._noisy(T=T)
            else:
                net = cpt._copy()
            new._add(net)
        return new

    def _parse_node_args2(self, Np, Nm=None):

        if Nm is None:
            cpt = self[Np]
            if isinstance(cpt, Node):
                Np, Nm = cpt.name, 0
            else:
                Np, Nm = cpt.node_names[0:2]
        return Np, Nm

    def _parse_node_args4(self, N1p, N1m, N2p, N2m, name):

        if N2p is None and N2m is None:

            arg1, arg2 = N1p, N1m

            if isinstance(arg1, tuple):
                N1p, N1m = arg1
                # TODO: check if there is a voltage source across these nodes
                # that will short out the applied source.
            else:
                try:
                    arg1 = self.elements[arg1]
                except:
                    pass
                if arg1 not in self.elements.values():
                    raise ValueError('Unknown component %s' % arg1)
                if name != 'as_ladder' and arg1.is_voltage_source:
                    # The killed voltage source will short the applied signal.
                    raise ValueError(
                        "Cannot determine transfer function across voltage source %s; you will need to remove it, e.g., new = cct.remove('%s')" % (arg1, arg1))
                N1p, N1m = [n.name for n in arg1.nodes[0:2]]

            if isinstance(arg2, tuple):
                N2p, N2m = arg2
            else:
                try:
                    arg2 = self.elements[arg2]
                except:
                    pass

                if arg2 not in self.elements.values():
                    raise ValueError('Unknown component %s' % arg2)
                N2p, N2m = [n.name for n in arg2.nodes[0:2]]

        elif N2p is None or N2m is None:
            raise ValueError('Expecting %s(cpt1, cpt2), %s(cpt1, (N2p, N2m), %s((N1p, N1m), cpt2), or %s(N1p, N1m, N2p, N2m)' % (
                name, name, name, name))

        return N1p, N1m, N2p, N2m

    def _potential_combine_names(self):

        names = []
        for name, elt in self.elements.items():
            if elt.type in ('V', 'I', 'R', 'NR', 'C', 'L', 'Y', 'Z'):
                names.append(name)
        return names

    def analyse(self):

        return Analysis(self, self.components)

    def augment_node_map(self, node_map=None):
        """Create a mapping dict for all nodes."""

        if node_map is None:
            node_map = {}

        # Add nodes with names like U1.vdd so they don't get renamed.
        ignore_nodes = []
        for node in self.nodes.keys():
            if node in node_map:
                continue
            parts = node.split('.')
            if len(parts) < 2:
                continue
            if parts[-2] in self._elements:
                node_map[node] = node
                ignore_nodes.append(node)

        # It would be desirable to renumber the nodes say from left to
        # right and top to bottom.  The schematic drawing algorithms
        # could help with this since they figure out the node
        # placement.

        enodes = self.equipotential_nodes

        # Rewrite the equipotential nodes removing nodes that need to
        # be ignored.
        enodes2 = {}
        for key, nodes in enodes.items():
            newnodes = []
            for node in nodes:
                if node not in ignore_nodes:
                    newnodes.append(node)

            if key in ignore_nodes:
                key = newnodes[0]
            enodes2[key] = newnodes
        enodes = enodes2

        if '0' in self.nodes and '0' not in node_map:
            node_map['0'] = '0'

        numbers = []
        for m in range(len(enodes)):
            numbers.append('%s' % (m + 1))

        # Check if user has supplied an unknown node in node_map.
        for old, new in node_map.items():
            if old not in self.nodes:
                raise ValueError('Unknown node %s' % old)
            if new in numbers:
                numbers.remove(new)

        # Ensure that enode keys are the nodes to be renamed.
        enodes2 = {}
        for key, nodes in enodes.items():
            newkey = None
            for node in nodes:
                if node not in node_map:
                    continue
                if newkey is not None:
                    # TODO, FIXME
                    raise ValueError(
                        'Cannot rename two nodes of same potential')
                newkey = node_map[node]
            if newkey is None:
                newkey = numbers.pop(0)
            enodes2[key] = (newkey, nodes)

        for key, foo in enodes2.items():

            newkey, nodes = foo

            snodes = sorted(nodes.copy())

            root = newkey
            snodes.remove(key)
            node_map[key] = root

            for m, enode in enumerate(snodes):
                node_map[enode] = root + '_%d' % (m + 1)

        return node_map

    def check(self):
        """Check if network contains a loop of voltage sources or a cut set of
        current sources."""

        return self.simplify(explain=True, modify=False)

    def circuit_graph(self):
        """Generate circuit graph for this netlist.   This is cached."""

        from .circuitgraph import CircuitGraph

        if hasattr(self, '_cg'):
            return self._cg

        self._cg = CircuitGraph.from_circuit(self)
        return self._cg

    def copy(self):
        """Create a copy of the netlist"""

        new = self._new()

        for cpt in self._elements.values():
            new._add(cpt._copy())
        return new

    def draw(self, filename=None, **kwargs):
        """Draw schematic of netlist.

        filename specifies the name of the file to produce.  If None,
        the schematic is displayed on the screen.

        Note, if using Jupyter, then need to first issue command %matplotlib inline

        kwargs include:
           label_ids: True to show component ids
           label_values: True to display component values
           draw_nodes: True to show all nodes, False to show no nodes,
             'primary' to show primary nodes,
             'connections' to show nodes that connect more than two components,
             'all' to show all nodes
           label_nodes: True to label all nodes, False to label no nodes,
             'primary' to label primary nodes,
             'alpha' to label nodes starting with a letter,
             'pins' to label nodes that are pins on a chip,
             'all' to label all nodes
           style: 'american', 'british', or 'european'
           scale: schematic scale factor, default 1.0
           node_spacing: spacing between component nodes, default 2.0
           cpt_size: size of a component, default 1.5
           dpi: dots per inch for png files
           help_lines: distance between lines in grid, default 0.0 (disabled)
           debug: True to display debug information
        """

        cct = self
        if kwargs.pop('s_model', False):
            cct = cct.s_model()

        return cct.sch.draw(filename=filename, **kwargs)

    def has(self, cpt):
        """Return True if cpt in elements."""

        return cpt in self.elements

    def in_parallel(self, cpt=None):
        """Return set of cpts in parallel with specified cpt.  If no cpt
        specified, return list of sets of parallel cpts."""

        if cpt is None:
            return self._in_parallel_all()

        if isinstance(cpt, Cpt):
            cpt = cpt.name

        names = self._in_parallel_names(cpt)
        if len(names) < 2:
            return set()
        names.discard(cpt)
        return list(names)

    def in_series(self, cpt=None):
        """Return set of cpts in series with specified cpt.  If no cpt
        specified, return list of sets of series cpts."""

        if cpt is None:
            return self._in_series_all()

        if isinstance(cpt, Cpt):
            cpt = cpt.name

        names = self._in_series_names(cpt)
        if len(names) < 2:
            return set()
        names.discard(cpt)
        return list(names)

    def kill_except(self, *args):
        """Return a new circuit with all but the specified sources killed;
        i.e., make the voltage sources short-circuits and the current
        sources open-circuits.  If no sources are specified, all
        independent sources (including initial conditions) are killed.

        """

        for arg in args:
            if arg not in self.independent_sources and arg != 'ICs':
                raise ValueError(
                    'Element %s is not a known independent source' % arg)
        sources = []
        for source in self.independent_sources:
            if source not in args:
                sources.append(source)
        if 'ICs' not in args:
            sources.append('ICs')

        return self._kill(sources)

    def kill(self, *args):
        """Return a new circuit with the specified sources killed; i.e., make
        the voltage sources short-circuits and the current sources
        open-circuits.  To kill initial conditions, specify `ICs'.  If
        no sources are specified, all independent sources (including
        initial conditions) are killed.

        """

        if len(args) == 0:
            return self.kill_except()

        sources = []
        for arg in args:
            if arg == 'ICs':
                sources.append(arg)
            elif arg in self.independent_sources:
                sources.append(arg)
            else:
                raise ValueError(
                    'Element %s is not a known independent source' % arg)

        return self._kill(sources)

    def kill_noise(self):
        """Return a new circuit with the independent noise voltage sources and
        noise current sources killed."""

        new = self._new()

        for cpt in self._elements.values():
            if cpt.independent_source and cpt.is_noisy:
                net = cpt._kill()
            else:
                net = cpt._copy()
            new._add(net)
        return new

    def kill_zero(self):
        """Return a new circuit with the independent zero voltage sources and
        zero current sources killed."""

        new = self._new()

        for cpt in self._elements.values():
            if (cpt.independent_source and
                (cpt.is_voltage_source and cpt.Voc == 0) or
                    (cpt.is_current_source and cpt.Isc == 0)):
                net = cpt._kill()
            else:
                net = cpt._copy()
            new._add(net)
        return new

    def match(self, pattern):
        """Return list of components names matching regular
        expression pattern."""

        from re import match

        elts = []
        for cpt in self._elements.values():
            if match(pattern, cpt.name):
                elts.append(cpt.name)
        return elts

    def netlist(self):
        """Return the current netlist."""

        return '\n'.join([str(cpt) for cpt in self._elements.values()])

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    def subs(self, *args):
        """Substitute symbols in the netlist.

        `args` is either:
        - two arguments, e.g. foo.subs(old, new)
        - one dict or set argument whose key/value items correspond to
          old/new pairs.

        For example, `b = a.subs({'R1': 1e3, 'R2': 9e3})`

        Note, this does not substitute component names.
        """

        if len(args) > 2:
            raise ValueError('Too many args')
        if len(args) == 2:
            subs_dict = {args[0]: args[1]}
        else:
            subs_dict = args[0]

        new = self._new()

        for cpt in self._elements.values():
            net = cpt._subs(subs_dict)
            new._add(net)
        return new

    def pre_initial_model(self):
        """Generate model for determining the pre-initial conditions."""

        new = self._new()

        for cpt in self._elements.values():
            net = cpt._pre_initial_model()
            new._add(net)
        return new

    def r_model(self):
        """"Create resistive equivalent model using companion circuits.
        This is experimental!"""

        new = self._new()

        for cpt in self._elements.values():
            net = cpt._r_model()
            new._add(net)
        return new

    def s_model(self, kind='s'):
        """"Create Laplace-domain model."""

        new = self._new()

        for cpt in self._elements.values():
            net = cpt._s_model(kind)
            new._add(net)
        return new

    def ss_model(self):
        """"Create state-space model by replacing inductors
        with current sources and capacitors with voltage sources."""

        new = self._new()

        for cpt in self._elements.values():
            net = cpt._ss_model()
            new._add(net)
        return new

    def state_space_model(self):
        """"Create state-space model by replacing inductors
        with current sources and capacitors with voltage sources."""

        return self.ss_model()

    def ac_model(self, var=omega):
        """"Create AC model for specified angular frequency (default
        omega)."""

        return self.s_model(j * var)

    def unconnected_nodes(self):
        """Return list of node names that are not connected."""

        return [node.name for node in self.nodes.values() if node.count <= 1]

    def Vname(self, name):
        return Vname(name, self.kind)

    def Iname(self, name):
        return Iname(name, self.kind)



"""This module provides support for inverse discrete-time Fourier
transforms.

 It calculates the inverse discrete-time Fourier transform using:

   x(n) = dt * \int_{-1 / (2 * dt)}^{1 / (2 * dt))} X(f) e^{j * 2 * \pi * n * dt * f} df

where dt is the sampling period.  The integral just needs to be
performed over any interval of length 1 / dt.

Copyright 2021--2022 Michael Hayes, UCECE

"""

import sympy as sym
from sympy import oo, DiracDelta
from .transformer import BilateralInverseTransformer
from .sym import sympify, AppliedUndef, j, pi
from .sym import dt
from .extrafunctions import UnitImpulse, UnitStep
from .utils import factor_const, scale_shift, remove_images
from .matrix import Matrix

__all__ = ('IDTFT', 'inverse_discrete_time_fourier_transform')


class IDTFTTransformer(BilateralInverseTransformer):

    name = 'inverse DTFT'

    def key(self, expr, f, n, **assumptions):
        return expr, f, n

    def noevaluate(self, expr, f, n):

        foo = expr * sym.exp(2 * j * pi * n * dt * f)
        result = dt * sym.Integral(foo, (f, -1 / (2 * dt), 1 / (2 * dt)))
        return result

    def check(self, expr, f, n, **assumptions):

        if expr.has(n):
            self.error('Expression depends on n')

    def rewrite(self, expr, var):
        """Remove images."""

        return remove_images(expr, var, dt)

    def sympy(self, expr, f, n):

        foo = expr * sym.exp(2 * j * pi * n * dt * f)
        result = dt * sym.integrate(foo, (f, -1 / (2 * dt), 1 / (2 * dt)))
        return result

    def func(self, expr, f, n):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], f)

        # Convert v(n) to V(f), etc.
        name = expr.func.__name__
        func = sym.Function(name[0].lower() + name[1:])

        result = func(n / scale) / abs(scale)

        if shift != 0:
            result = result * sym.exp(-2 * sym.I * sym.pi * f * shift / scale)

        return result

    def function(self, expr, f, n):

        # Handle expressions with a function of FOO, e.g.,
        # v(n), v(n) * y(n),  3 * v(n) / n, v(4 * a * n), etc.,

        if not expr.has(AppliedUndef):
            self.error()

        const, expr = factor_const(expr, f)

        if isinstance(expr, AppliedUndef):
            return self.func(expr, f, n) * const

        nsym = sympify(str(n))
        expr = expr.subs(nsym, n)

        rest = sym.S.One
        undefs = []
        for factor in expr.as_ordered_factors():
            if isinstance(factor, AppliedUndef):
                if factor.args[0] != f:
                    self.error('Weird function %s not of %s' % (factor, f))
                undefs.append(factor)
            else:
                rest *= factor

        if rest.has(AppliedUndef):
            # Have something like 1/V(f)
            self.error()

        exprs = undefs
        if rest.has(n):
            exprs = exprs + [rest]
            rest = sym.S.One

        result = self.term(exprs[0], f, n) * rest

        if len(exprs) == 1:
            return result * const

        self.error('TODO')

    def term(self, expr, f, n):

        const, expr = factor_const(expr, f)

        # Check for constant.
        if not expr.has(f):
            return expr * UnitImpulse(n) * const

        if expr.has(AppliedUndef):
            # Handle v(n), v(n) * y(n), 3 * v(n) / n etc.
            return self.function(expr, f, n) * const

        # This handles DiracDelta the long way but could head off at pass...
        return const * self.sympy(expr, f, n)


idtft_transformer = IDTFTTransformer()


def inverse_discrete_time_fourier_transform(expr, f, n, evaluate=True,
                                            **assumptions):
    """Compute bilateral inverse discrete-time Fourier transform of expr.

    Undefined functions such as X(f) are converted to x(n).

    """

    return idtft_transformer.transform(expr, f, n, evaluate=evaluate,
                                       **assumptions)


def IDTFT(expr, f, n, evaluate=True, **assumptions):
    """Compute bilateral inverse discrete-time Fourier transform of expr.

    Undefined functions such as X(f) are converted to x(n).
    """

    return idtft_transformer.transform(expr, f, n, evaluate=evaluate,
                                        **assumptions)



from lcapy import f, sqrt
Vn = 1e-9 * sqrt(3.5 / f + 1)
ax = (Vn * 1e9).plot((0.1, 10e3), loglog=True)
ax.grid(True, 'both')
ax.set_ylabel('Noise voltage density nV$/\sqrt{\mathrm{Hz}}$')
ax.set_ylim(0.1, 10)

from matplotlib.pyplot import savefig
savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



from lcapy import s, j, pi, f, transfer, j2pif
from matplotlib.pyplot import savefig
from numpy import logspace

# Note, this has a marginally stable impulse response since it has a
# pole at s = 0.
H = transfer((s - 2) * (s + 3) / (s * (s - 2 * j) * (s + 2 * j)))

fv = logspace(-1, 3, 400)
H(j2pif).dB.plot(fv, log_scale=True)

savefig('tf1-bode-plot.png')



"""This module provides the NetlistOpsMixin class.

Copyright 2023 Michael Hayes, UCECE

"""

# TODO: Optimise oneport in a similar manner to ladder.  This will avoid
# solving for all nodal voltages.  Voc, ISc, impedance, and admittance
# can then be sped up using the oneport.

from .admittance import admittance
from .current import current, current_sign
from .impedance import impedance
from .matrix import Matrix
from .statespace import StateSpace
from .symbols import s
from .transfer import transfer
from warnings import warn


class NetlistOpsMixin:

    def admittance(self, Np, Nm=None):
        """Return driving-point Laplace-domain admittance between nodes
        Np and Nm with independent sources killed and initial
        conditions ignored."""

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)

        new = self.kill()
        new._add_ground(Nm)
        test = new._add_test_voltage_source(Np, Nm)
        If = current_sign(-new[test].I, True)
        return admittance(If.laplace().sympy)

    def conductance(self, Np, Nm=None):
        """Return conductance (inverse resistance) between nodes Np and Nm
        with independent sources killed.  The result is in the AC (omega)
        domain.    See also resistance, reactance, susceptance."""
        return self.impedance(Np, Nm).G

    def current_gain(self, N1p, N1m, N2p=None, N2m=None):
        """Create Laplace-domain current transfer function I2(s) / I1(s) where:
        I1 is the test current applied between N1p and N1m
        I2 is the measured short-circuit current flowing from N2m to N2p

        Note, the currents are considered to be flowing into the
        positive nodes as is the convention with two-ports.  Thus the
        input and output currents have opposite directions and so a
        piece of wire has a current gain of -1.

        Note, independent sources are killed and initial conditions
        are ignored.

        Alternative forms are:
            current_gain(N1p, N1m, N2p, N2m)
            current_gain(cpt1, cpt2)
            current_gain((N1p, N1m), cpt2)
            current_gain(cpt1, (N2p, N2m))

        """

        ladder = self._ladder(N1p, N1m, N2p, N2m)
        if ladder is not None:
            return ladder.current_gain

        N1p, N1m, N2p, N2m = self._parse_node_args4(N1p, N1m, N2p, N2m,
                                                    'current_gain')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        new = self.apply_test_current_source(N1p, N1m)
        H = transfer(-new.Isc(N2p, N2m).laplace())
        H.causal = True
        return H

    def impedance(self, Np, Nm=None):
        """Return driving-point Laplace-domain impedance between nodes
        Np and Nm with independent sources killed and initial
        conditions ignored."""

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)

        new = self.apply_test_current_source(Np, Nm)
        Vf = new.Voc(Np, Nm)
        return impedance(Vf.laplace().sympy)

    def Isc(self, Np, Nm=None, **kwargs):
        """Return short-circuit transform-domain current between nodes Np and
        Nm."""

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)

        new = self.copy()
        if new.is_causal:
            new.add('Vshort_ %s %s step 0' % (Np, Nm))
        else:
            new.add('Vshort_ %s %s 0' % (Np, Nm))

        Isc = current_sign(new.get_I('Vshort_', **kwargs), True)

        new.remove('Vshort_')

        return Isc

    def isc(self, Np, Nm=None):
        """Return short-circuit t-domain current between nodes Np and Nm."""

        return self.Isc(Np, Nm).time()

    def _ladder(self, N1p, N1m, N2p=None, N2m=None):

        # TODO: determine point at which it is more efficient to
        # find ladder network
        if len(self.elements) < 6:
            return None
        return self.kill().ladder(N1p, N1m, N2p, N2m)

    def ladder(self, N1p, N1m, N2p=None, N2m=None):
        """Return two-port unbalanced ladder network or `None` if the netlist
        does not have a ladder topology between the specified nodes.

        The input port is defined by the nodes `N1p` and `N1m`.
        The output port is defined by the nodes `N2p` and `N2m`.

        The nodes `N1p` and `N1m` must be the same.

        Alternative forms are:
            ladder(N1p, N1m, N2p, N2m)
            ladder(cpt1, cpt2)
            ladder((N1p, N1m), cpt2)
            ladder(cpt1, (N2p, N2m))

        This method can be used to generate a transfer function, for example:

        `cct.ladder(1, 0, 10, 0).voltage_gain`

        This can be much faster than using `cct.voltage_gain(1, 0, 10, 0)`
        for circuits with a large ladder topology, since the latter method needs
        to determine all the node voltages.

        """

        from lcapy.laddernetworkmaker import LadderNetworkMaker

        N1p, N1m, N2p, N2m = self._parse_node_args4(N1p, N1m, N2p, N2m,
                                                    'ladder')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        lm = LadderNetworkMaker(self)

        tp = lm.make(N1p, N1m, N2p, N2m)
        return tp

    def norton(self, Np, Nm=None):
        """Return Laplace-domain Norton model between nodes Np and Nm.

        If Np is a component name, create model using its component nodes."""

        from .oneport import I, Y

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)
        Isc = self.Isc(Np, Nm)
        Ysc = self.admittance(Np, Nm)

        return (I(Isc) | Y(Ysc)).simplify()

    def oneport(self, Np, Nm=None):
        """Return oneport object between nodes Np and Nm.  This might be a
        Thevenin network, a Norton network, or a single component.

        If Np is a component name, create model using its component nodes."""

        try:
            return self.norton(Np, Nm)
        except:
            return self.thevenin(Np, Nm)

    def reactance(self, Np, Nm=None):
        """Return reactance between nodes Np and Nm with independent
        sources killed.  The result is in the AC (omega) domain.
        See also conductance, resistance, susceptance."""
        return self.impedance(Np, Nm).X

    def resistance(self, Np, Nm=None):
        """Return resistance between nodes Np and Nm with independent
        sources killed.  The result is in the AC (omega) domain.
        See also conductance, reactance, susceptance."""
        return self.impedance(Np, Nm).R

    def state_space(self, node_voltages=None, branch_currents=None):
        """Generate state-space representation.

        `node_voltages` is a list of node names to use as voltage outputs.
        If `None` use all the unique node names.

        `branch_currents` is a list of component names to use as
        current outputs.  If `None` use all the components.

        Here's an example:
        `cct = Circuit('cct.sch')
        ss = cct.state_space(node_voltages=['1', '3'], branch_currents=['L1', 'L2'])`
        """

        ss = StateSpace.from_circuit(self, node_voltages, branch_currents)
        return ss

    def susceptance(self, Np, Nm=None):
        """Return susceptance (inverse reactance) between nodes Np and Nm with
        independent sources killed.  The result is in the AC (omega)
        domain.  See also conductance, reactance, resistance."""
        return self.impedance(Np, Nm).B

    def thevenin(self, Np, Nm=None):
        """Return Laplace-domain Thevenin oneport model between nodes Np and Nm.

        If Np is a component name, create model using its component nodes."""

        from .oneport import V, Z

        Np, Nm = self._parse_node_args2(Np, Nm)
        Np, Nm = self._check_nodes(Np, Nm)
        Voc = self.Voc(Np, Nm)
        Zoc = self.impedance(Np, Nm)

        return (V(Voc) + Z(Zoc)).simplify()

    def transfer(self, N1p, N1m, N2p=None, N2m=None):
        """Create Laplace-domain voltage transfer function V2(s) / V1(s) where:
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        Note, independent sources are killed and initial conditions
        are ignored.

        Alternative forms are:
            transfer(N1p, N1m, N2p, N2m)
            transfer(cpt1, cpt2)
            transfer((N1p, N1m), cpt2)
            transfer(cpt1, (N2p, N2m))
        """

        ladder = self._ladder(N1p, N1m, N2p, N2m)
        if ladder is not None:
            return ladder.voltage_gain

        N1p, N1m, N2p, N2m = self._parse_node_args4(N1p, N1m, N2p, N2m,
                                                    'transfer')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        new = self.apply_test_voltage_source(N1p, N1m)
        V2 = new.Voc(N2p, N2m)
        H = transfer(V2.laplace())
        H.causal = True
        return H

    def twoport(self, N1p, N1m, N2p=None, N2m=None, model='B'):
        """Create Laplace-domain twoport model for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        `model` is `A, `B`, `G`, `H`, `Y`, or `Z`.

        Alternative forms are:
            twoport(N1p, N1m, N2p, N2m)
            twoport(cpt1, cpt2)
            twoport((N1p, N1m), cpt2)
            twoport(cpt1, (N2p, N2m))
        """

        from .twoport import TwoPortAModel, TwoPortBModel, TwoPortGModel
        from .twoport import TwoPortHModel, TwoPortYModel, TwoPortZModel

        N1p, N1m, N2p, N2m = self._parse_node_args4(
            N1p, N1m, N2p, N2m, 'twoport')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        # TODO, generalise for not just Laplace-domain.

        new = self.copy()
        new._add_ground(N1m)

        if model == 'A':
            V1a = new.Voc(N1p, N1m, nowarn=True)(s)
            I1a = new.Isc(N1p, N1m, nowarn=True)(s)
            A = new.Aparams(N1p, N1m, N2p, N2m)
            return TwoPortAModel(A, V1a=V1a, I1a=I1a)
        elif model == 'B':
            V2b = new.Voc(N2p, N2m, nowarn=True)(s)
            I2b = new.Isc(N2p, N2m, nowarn=True)(s)
            A = new.Aparams(N1p, N1m, N2p, N2m)
            return TwoPortBModel(A.Bparams, V2b=V2b, I2b=I2b)
        elif model == 'Z':
            V1 = new.Voc(N1p, N1m, nowarn=True)(s)
            V2 = new.Voc(N2p, N2m, nowarn=True)(s)
            Z = new.Zparams(N1p, N1m, N2p, N2m)
            return TwoPortZModel(Z, V1z=V1, V2z=V2)
        elif model == 'Y':
            I1 = new.Isc(N1p, N1m, nowarn=True)(s)
            I2 = new.Isc(N2p, N2m, nowarn=True)(s)
            Z = new.Zparams(N1p, N1m, N2p, N2m)
            return TwoPortYModel(Z.Yparams, I1y=I1, I2y=I2)
        elif model == 'G':
            I1 = new.Isc(N1p, N1m, nowarn=True)(s)
            V2 = new.Voc(N2p, N2m, nowarn=True)(s)
            Z = new.Zparams(N1p, N1m, N2p, N2m)
            return TwoPortGModel(Z.Gparams, I1g=I1, V2g=V2)
        elif model == 'H':
            V1 = new.Voc(N1p, N1m, nowarn=True)(s)
            I2 = new.Isc(N2p, N2m, nowarn=True)(s)
            Z = new.Zparams(N1p, N1m, N2p, N2m)
            return TwoPortHModel(Z.Hparams, V1h=V1, I2h=I2)
        else:
            raise ValueError('Model %s unknown, must be B, H, Y, or Z' % model)

    def transadmittance(self, N1p, N1m, N2p=None, N2m=None):
        """Create Laplace-domain transadmittance (transfer admittance) function
        I2(s) / V1(s) where:
          V1 is the test voltage applied between N1p and N1m
          I2 is the measured short-circuit current flowing from N2m to N2p.

        Note, I2 is considered to be flowing into the positive node as
        is the convention with two-ports.  Thus the transadmittance of
        a series resistor with resistance R is -1 / R.

        Note, independent sources are killed and initial conditions
        are ignored.

        Alternative forms are:
            transadmittance(N1p, N1m, N2p, N2m)
            transadmittance(cpt1, cpt2)
            transadmittance((N1p, N1m), cpt2)
            transadmittance(cpt1, (N2p, N2m))

        """

        ladder = self._ladder(N1p, N1m, N2p, N2m)
        if ladder is not None:
            return ladder.transadmittance

        N1p, N1m, N2p, N2m = self._parse_node_args4(N1p, N1m, N2p, N2m,
                                                    'transadmittance')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        new = self.apply_test_voltage_source(N1p, N1m)
        H = admittance(-new.Isc(N2p, N2m).laplace())
        H.causal = True
        return H

    def transimpedance(self, N1p, N1m, N2p=None, N2m=None):
        """Create Laplace-domain transimpedance (transfer impedance) function
        V2(s) / I1(s) where:
          I1 is the test current applied between N1p and N1m
          V2 is the measured open-circuit voltage between N2p and N2m.

        Note, I1 is considered to be flowing into the positive node as
        is the convention with two-ports.

        Note, independent sources are killed and initial conditions
        are ignored.

        Alternative forms are:
            transimpedance(N1p, N1m, N2p, N2m)
            transimpedance(cpt1, cpt2)
            transimpedance((N1p, N1m), cpt2)
            transimpedance(cpt1, (N2p, N2m))
        """

        ladder = self._ladder(N1p, N1m, N2p, N2m)
        if ladder is not None:
            return ladder.transimpedance

        N1p, N1m, N2p, N2m = self._parse_node_args4(N1p, N1m, N2p, N2m,
                                                    'transadmittance')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        new = self.apply_test_current_source(N1p, N1m)
        H = impedance(new.Voc(N2p, N2m).laplace())
        H.causal = True
        return H

    def Voc(self, Np, Nm=None, **kwargs):
        """Return open-circuit transform-domain voltage between nodes Np and
        Nm."""

        return self._get_Vd(Np, Nm, **kwargs)

    def voc(self, Np, Nm=None):
        """Return open-circuit t-domain voltage between nodes Np and Nm."""

        return self.Voc(Np, Nm).time()

    def voltage_gain(self, N1p, N1m, N2p=None, N2m=None):
        """Create Laplace-domain voltage transfer function V2(s) / V1(s) where:
        V1 is the test voltage applied between N1p and N1m
        V2 is the measured open-circuit voltage between N2p and N2m

        Note, independent sources are killed and initial conditions
        are ignored.

        Alternative forms are:
            voltage_gain(N1p, N1m, N2p, N2m)
            voltage_gain(cpt1, cpt2)
            voltage_gain((N1p, N1m), cpt2)
            voltage_gain(cpt1, (N2p, N2m))
        """

        ladder = self._ladder(N1p, N1m, N2p, N2m)
        if ladder is not None:
            return ladder.voltage_gain

        # This is the same as transfer.
        N1p, N1m, N2p, N2m = self._parse_node_args4(N1p, N1m, N2p, N2m,
                                                    'voltage_gain')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)

        new = self.apply_test_voltage_source(N1p, N1m)
        V2 = new.Voc(N2p, N2m)
        H = transfer(V2.laplace())
        H.causal = True
        return H

    def Aparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create A-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Bparams, Gparams, Hparams, Sparams, Tparams, Yparams, and Zparams.
        """

        from .twoport import AMatrix

        N1p, N1m, N2p, N2m = self._parse_node_args4(
            N1p, N1m, N2p, N2m, 'Aparams')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)
        new = self.kill()
        new._add_ground(N1m)

        try:
            test = new._add_test_voltage_source(N1p, N1m)

            # A11 = V1 / V2 with I2 = 0
            # Apply V1 and measure V2 with port 2 open-circuit
            A11 = new.Voc(N1p, N1m)(s) / new.Voc(N2p, N2m)(s)

            # A12 = V1 / I2 with V2 = 0
            # Apply V1 and measure -I2 with port 2 short-circuit
            A12 = new.Voc(N1p, N1m)(s) / new.Isc(N2p, N2m)(s)

            new.remove(test)

            test = new._add_test_current_source(N1p, N1m)

            # A21 = I1 / V2 with I2 = 0
            # Apply I1 and measure V2 with port 2 open-circuit
            try:
                A21 = current(0 * s + 1) / new.Voc(N2p, N2m)(s)
            except ValueError:
                # It is likely there is an open-circuit.
                new2 = new.copy()
                new2.add('W %s %s' % (N2p, N2m))
                A21 = -new2[test].I(s) / new2.Voc(N2p, N2m)(s)
                A21 = 0

            # A22 = I1 / I2 with V2 = 0
            # Apply I1 and measure -I2 with port 2 short-circuit
            A22 = current(0 * s + 1) / new.Isc(N2p, N2m)(s)

            new.remove(test)
            A = AMatrix(((A11, A12), (A21, A22)))
            return A

        except ValueError:
            warn('Cannot create A matrix directly; trying via Z matrix')
            Z = self.Zparams(N1p, N1m, N2p, N2m)
            return Z.Aparams

    def Bparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create B-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Gparams, Hparams, Sparams, Tparams, Yparams, and Zparams.
        """
        return self.Aparams(N1p, N1m, N2p, N2m).Bparams

    def Gparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create G-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Bparams, Hparams, Sparams, Tparams, Yparams, and Zparams.
        """
        return self.Aparams(N1p, N1m, N2p, N2m).Gparams

    def Hparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create H-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Bparams, Gparams, Sparams, Tparams, Yparams, and Zparams.
        """
        return self.Aparams(N1p, N1m, N2p, N2m).Hparams

    def Sparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create S-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Bparams, Gparams, Hparams, Tparams, Yparams, and Zparams.
        """
        return self.Aparams(N1p, N1m, N2p, N2m).Sparams

    def Tparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create T-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Bparams, Gparams, Hparams, Sparams, Yparams, and Zparams.
        """
        return self.Tparams(N1p, N1m, N2p, N2m).Hparams

    def Yparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create Y-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Bparams, Gparams, Hparams, Sparams, Tparams, and Zparams.
        """
        return self.Zparams(N1p, N1m, N2p, N2m).Yparams

    def Zparams(self, N1p, N1m, N2p=None, N2m=None):
        """Create Z-parameters for two-port defined by nodes N1p, N1m, N2p, and N2m, where:
        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]

        See also twoport, Aparams, Bparams, Gparams, Hparams, Sparams, Tparams, and Yparams.
        """
        from .twoport import ZMatrix

        # TODO, generalise to multiports.

        N1p, N1m, N2p, N2m = self._parse_node_args4(
            N1p, N1m, N2p, N2m, 'Zparams')
        N1p, N1m, N2p, N2m = self._check_nodes(N1p, N1m, N2p, N2m)
        new = self.kill()
        new._add_ground(N1m)

        try:
            test = new._add_test_current_source(N1p, N1m)

            # Z11 = V1 / I1 with I2 = 0
            # Apply I1 and measure V1 with port 2 open-circuit
            Z11 = impedance(new.Voc(N1p, N1m)(s))

            # Z21 = V2 / I1 with I2 = 0
            # Apply I1 and measure V2 with port 2 open-circuit
            Z21 = impedance(new.Voc(N2p, N2m)(s))

            new.remove(test)

            test = new._add_test_current_source(N2p, N2m)

            # Z12 = V1 / I2 with I1 = 0
            # Apply I2 and measure V1 with port 1 open-circuit
            Z12 = impedance(new.Voc(N1p, N1m)(s))

            # Z22 = V2 / I2 with I1 = 0
            # Apply I2 and measure V2 with port 1 open-circuit
            Z22 = impedance(new.Voc(N2p, N2m)(s))

            new.remove(test)

            Z = ZMatrix(((Z11, Z12), (Z21, Z22)))
            return Z

        except ValueError as e:
            raise ValueError('Cannot create Z matrix: %s' % e)

    def Yparamsn(self, *nodes):
        """Create Y-parameters for N-port defined by list of node-pairs.

        See also Yparams for a two port.

        """

        nodes = self._check_nodes(*nodes)
        if len(nodes) % 2 == 1:
            raise ValueError('Need an even number of nodes.')
        ports = []
        for m in range(len(nodes) // 2):
            ports.append((nodes[m * 2], nodes[m * 2 + 1]))

        new = self.kill()
        new._add_ground(nodes[1])

        try:

            Y = Matrix.zeros(len(ports))

            for col in range(len(ports)):

                for row in range(len(ports)):
                    if row == col:
                        new.add('V%d_ %s %s {DiracDelta(t)}' % (
                            row, ports[row][0], ports[row][1]))
                    else:
                        new.add('V%d_ %s %s 0' %
                                (row, ports[row][0], ports[row][1]))

                for row in range(len(ports)):
                    Y[row, col] = admittance(new.elements['V%d_' % row].I(s))

                for row in range(len(ports)):
                    new.remove('V%d_' % row)
            return Y

        except ValueError as e:
            raise ValueError('Cannot create Y matrix: %s' % e)

    def Yparams3(self, N1p, N1m, N2p, N2m, N3p, N3m):
        """Create Y-parameters for three-port defined by nodes N1p, N1m, N2p,
        N2m, N3p, and N3m, where:

        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        I3 is the current flowing into N3p and out of N3m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]
        V3 is V[N3p] - V[N3m]

        See also Yparams for a two port and Yparamsn for an N-port.

        """

        return self.Yparamsn(N1p, N1m, N2p, N2m, N3p, N3m)

    def Zparamsn(self, *nodes):
        """Create Z-parameters for N-port defined by list of node-pairs.

        See also Zparams for a two port.

        """

        nodes = self._check_nodes(*nodes)
        if len(nodes) % 2 == 1:
            raise ValueError('Need an even number of nodes.')
        ports = []
        for m in range(len(nodes) // 2):
            ports.append((nodes[m * 2], nodes[m * 2 + 1]))

        new = self.kill()
        new._add_ground(nodes[1])

        try:

            Z = Matrix.zeros(len(ports))

            for col in range(len(ports)):
                new.add('I_ %s %s {DiracDelta(t)}' %
                        (ports[col][0], ports[col][1]))

                for row in range(len(ports)):
                    Z[row, col] = impedance(
                        new.Voc(ports[row][0], ports[row][1])(s))

                new.remove('I_')
            return Z

        except ValueError as e:
            raise ValueError('Cannot create Z matrix: %s' % e)

    def Zparams3(self, N1p, N1m, N2p, N2m, N3p, N3m):
        """Create Z-parameters for three-port defined by nodes N1p, N1m, N2p,
        N2m, N3p, and N3m, where:

        I1 is the current flowing into N1p and out of N1m
        I2 is the current flowing into N2p and out of N2m
        I3 is the current flowing into N3p and out of N3m
        V1 is V[N1p] - V[N1m]
        V2 is V[N2p] - V[N2m]
        V3 is V[N3p] - V[N3m]

        See also Zparams for a two port and Zparamsn for an N-port.

        """

        return self.Zparamsn(N1p, N1m, N2p, N2m, N3p, N3m)



from lcapy import *
import numpy as np
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy discrete-time state space analysis
    """

    def assertEqual2(self, ans1, ans2, comment):

        ans1 = ans1.canonical()
        ans2 = ans2.canonical()

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_transfer(self):
        Z = (s**2 + 3) / (s**3 + 2 * s + 10)

        ss = Z.state_space()

        self.assertEqual(ss.Nx, 3, "Nx")
        self.assertEqual(ss.Ny, 1, "Ny")
        self.assertEqual(ss.Nu, 1, "Nu")
        self.assertEqual(ss.is_stable, False, "is_stable")
        self.assertEqual(ss.is_controllable, True, "is_controllable")
        self.assertEqual(ss.is_observable, True, "is_observable")
        self.assertEqual(ss.is_symbolic, False, "is_symbolic")
        self.assertEqual(ss.G[0], Z, "G")

        sso = Z.state_space('OCF')
        self.assertEqual(sso.G[0], Z, "G")

        ssd = Z.state_space('DCF')
        #self.assertEqual(ssd.G[0], Z, "G")

    def test_balanced(self):

        A = [[1, -2], [3, -4]]
        B = [5, 7]
        C = [[6, 8]]
        D = [9]

        ss = StateSpace(A, B, C, D)

        ssb = ss.balance()

        h = ss.hankel_singular_values

        H = np.diag(h.numpy.squeeze())

        self.assertEqual(np.allclose(
            ssb.Wo.numpy, ssb.Wc.numpy), True, "Wo==Wc")
        self.assertEqual(np.allclose(ssb.Wo.numpy, H),
                         True, "Hankel singular values")
        self.assertEqual(ss.eigenvalues, [-1, -2], "eigen values")

    def transfer(self):

        ss = DTStateSpace(((0, 1), (1, 1)), (1, 1), ((1, 2),), [1])

        umin = ss.minimum_energy_input(2, [5, 7], [0, 0])
        self.assertEqual(umin, [2, 3], "minimum_energy_input")

        umin = ss.minimum_energy(2, [5, 7], [0, 0])
        self.assertEqual(umin, 13, "minimum_energy")

        xfinal = ss.state_transfer([[2], [3]], xinitial=[0, 0])
        self.assertEqual(xfinal, [5, 7], "state_transfer")



from lcapy.domains import domains
from lcapy.quantities import quantities


units_mapping = {
    '': 'S.One',
    'V': 'uu.volt', 'A': 'uu.ampere',
    'V/Hz': 'uu.volt / uu.Hz',
    'A/Hz': 'uu.ampere / uu.Hz',
    'V/sqrt(Hz)': 'uu.volt / sqrt(uu.Hz)',
    'A/sqrt(Hz)': 'uu.ampere / sqrt(uu.Hz)',
    'ohm': 'uu.ohm',
    'S': 'uu.siemens',
    'ohm/s': 'uu.ohm / uu.s',
    'S/s': 'uu.siemens / uu.s',
    'ohm^2/s^2': '(uu.ohm / uu.s)**2',
    'S^2/s^2': '(uu.siemens / uu.s)**2',
    'V^2': 'uu.volt**2',
    'A^2': 'uu.ampere**2',
    'V^2/Hz^2': '(uu.volt / uu.Hz)**2',
    'A^2/Hz^2': '(uu.ampere / uu.Hz)**2',
    'ohm^2': 'uu.ohm**2', 'S^2': 'uu.siemens**2',
    'W': 'uu.watt', '/s': '1 / uu.s'}


class ClassesMake:

    def __init__(self, domains, quantities, kind):

        self.domains = domains
        self.quantities = quantities
        self.classes = []
        self.all = []
        self.cdict = {}
        # expr or seq
        self.kind = kind

        self.class_suffix = 'Sequence' if kind == 'seq' else ''
        self.baseclass_suffix = 'Sequence' if kind == 'seq' else 'Expression'

    def make(self):

        for domain in self.domains.values():

            if domain.is_superposition_domain:
                continue

            self.cdict[domain.domain] = {}

            self.cdict[domain.domain]['undefined'] = domain.__name__ + \
                self.baseclass_suffix

            for quantity in self.quantities.values():

                class_maker = ClassMake(domain, quantity)

                name, s = class_maker.make(self.class_suffix,
                                           self.baseclass_suffix)
                self.all.append(name)
                self.classes.append(s)

                self.cdict[domain.domain][quantity.quantity] = name

    def write(self):

        filename = self.kind + 'classes.py'
        var = self.kind + 'classes'

        s = '''
# This file is automatically generated by makeclasses.py

from .units import units, u as uu
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin
from .voltagesquaredmixin import VoltageSquaredMixin
from .currentsquaredmixin import CurrentSquaredMixin
from .admittancesquaredmixin import AdmittanceSquaredMixin
from .impedancesquaredmixin import ImpedanceSquaredMixin
from .powermixin import PowerMixin
from sympy import sqrt, S'''

        if self.kind == 'expr':
            s += '''
from .cexpr import ConstantDomainExpression
from .cexpr import ConstantTimeDomainExpression
from .cexpr import ConstantFrequencyResponseDomainExpression
from .fexpr import FourierDomainExpression
from .jfexpr import FrequencyResponseDomainExpression
from .jomegaexpr import AngularFrequencyResponseDomainExpression
from .kexpr import DiscreteFourierDomainExpression
from .nexpr import DiscreteTimeDomainExpression
from .noisefexpr import FourierNoiseDomainExpression
from .noiseomegaexpr import AngularFourierNoiseDomainExpression
from .normfexpr import NormFourierDomainExpression
from .normomegaexpr import NormAngularFourierDomainExpression
from .omegaexpr import AngularFourierDomainExpression
from .phasor import PhasorDomainExpression, PhasorRatioDomainExpression
from .sexpr import LaplaceDomainExpression
from .texpr import TimeDomainExpression
from .uexpr import UndefinedDomainExpression
from .zexpr import ZDomainExpression
'''

        if self.kind == 'seq':
            s += '''
from .kseq import DiscreteFourierDomainSequence
from .nseq import DiscreteTimeDomainSequence
from .zseq import ZDomainSequence
'''

        s += '\n__all__ = (' + ', '.join(["'" + name +
                                          "'" for name in self.all]) + ')' + '\n'

        s += ''.join(self.classes)

        s += '\n\n%s = {' % var

        for d, dv in self.cdict.items():
            s += "'" + d + "': {"
            for q, qv in dv.items():
                s += "'" + q + "': " + qv + ',\n'
            s += '}\n,'
        s += '}\n'

        with open(filename, 'wt') as outfile:
            print(s, file=outfile)


class ClassMake:

    def __init__(self, domain, quantity):

        self.domain = domain
        self.quantity = quantity

    def make(self, class_suffix, baseclass_suffix):

        quantity = self.quantity
        domain = self.domain

        quantityunits = quantity.quantity_units

        unitsstring = quantityunits
        if quantity.quantity in ('voltage', 'current'):
            if (domain.is_laplace_domain or
                domain.is_fourier_domain or
                    domain.is_angular_fourier_domain):
                unitsstring = '%s/Hz' % quantityunits
        elif quantity.quantity in ('voltagesquared', 'currentsquared'):
            if (domain.is_laplace_domain or
                domain.is_fourier_domain or
                    domain.is_angular_fourier_domain):
                unitsstring = '%s/Hz^2' % quantityunits
        elif quantity.quantity in ('impedance', 'admittance', 'transfer'):
            if domain.is_time_domain:
                unitsstring = '%s/s' % quantityunits
        elif quantity.quantity in ('admittancesquared', 'impedancesquared'):
            if domain.is_time_domain:
                unitsstring = '%s/s^2' % quantityunits

        classname = domain.__name__ + quantity.quantity.capitalize() + class_suffix
        classname = classname.replace('Transfer', 'TransferFunction')

        baseclassname = domain.__name__ + baseclass_suffix

        # TODO: perhaps rename transfer function in time domain to
        # impulse response?
        docstring = '%s-domain %s (units %s).' % (domain.domain_label,
                                                  quantity.quantity,
                                                  unitsstring)

        s = '''


class %s(%s, %s):
    """%s"""

    _default_units = %s''' % (classname, quantity.__name__, baseclassname,
                              docstring,  units_mapping[unitsstring])

        return classname, s


expr_maker = ClassesMake(domains, quantities, 'expr')
expr_maker.make()
expr_maker.write()

seq_domains = {}
for domain in ['discrete time', 'discrete fourier', 'Z']:
    seq_domains[domain] = domains[domain]

seq_maker = ClassesMake(seq_domains, quantities, 'seq')
seq_maker.make()
seq_maker.write()



"""This module wraps SymPy functions and provides a few others.
The underlying SymPy function is obtained with the sympy attribute.

Copyright 2014--2023 Michael Hayes, UCECE
"""

from .units import u as uu
import sympy as sym

__all__ = ('conjugate', 'sqrt', 'exp', 'log', 'log10', 'sin', 'cos',
           'tan', 'cot', 'asin', 'acos', 'atan', 'atan2', 'acot',
           'sinh', 'cosh', 'tanh', 'coth', 'asinh', 'acosh', 'atanh',
           'acoth', 'gcd', 'sign', 'diff', 'integrate', 'u', 'H',
           'heaviside', 'Heaviside', 'delta', 'DiracDelta', 'ui',
           'us', 'unitimpulse', 'unitstep', 'UnitImpulse', 'UnitStep',
           'rect', 'sinc', 'sincn', 'sincu', 'psinc', 'tri', 'trap',
           'ramp', 'rampstep', 'Sum', 'dtrect', 'dtsign', 'Piecewise',
           'Eq', 'Ne', 'Lt', 'Le', 'Gt', 'Ge', 'Derivative',
           'Integral', 'Max', 'Min', 're', 'im', 'MatMul', 'MatAdd',
           'degrees', 'radians', 'limit', 'besseli', 'besselj')

from .extrafunctions import Degrees as Degrees1
from .extrafunctions import Radians as Radians1
from .extrafunctions import UnitImpulse as UnitImpulse1
from .extrafunctions import UnitStep as UnitStep1
from .extrafunctions import sincn as sincn1
from .extrafunctions import sincu as sincu1
from .extrafunctions import psinc as psinc1
from .extrafunctions import rect as rect1
from .extrafunctions import dtrect as dtrect1
from .extrafunctions import dtsign as dtsign1
from .extrafunctions import tri as tri1
from .extrafunctions import trap as trap1
from .extrafunctions import ramp as ramp1
from .extrafunctions import rampstep as rampstep1


class Function(object):

    def __init__(self, arg):

        if isinstance(arg, str):
            # Handled AppliedUndef
            arg = sym.Function(arg)
        self.expr = arg

    def __repr__(self):

        return repr(self.expr)

    @property
    def sympy(self):
        """Return SymPy function."""

        return self.expr

    def __call__(self, *args, **kwargs):

        func = self.expr

        e_args = list(map(expr, args))

        if func == sym.Piecewise:
            cls = e_args[0][0].__class__
        elif func in (sym.Integral, sym.integrate):
            # Integral(const, (tau, 0, t)) -> use t class
            # Integral(integrand, (tau, a, b)) -> use integrand class
            if e_args[0].is_constant:
                cls = e_args[1][2].__class__
            else:
                cls = e_args[0].__class__
        else:
            cls = e_args[0].__class__

            # Handle cases like atan2(1, omega)
            if len(e_args) > 1:
                if e_args[0].is_constant:
                    cls = e_args[1].__class__

        try:
            if cls.is_discrete_time_domain or cls.is_discrete_fourier_domain:

                # TODO: fix for Heaviside change
                for old, new in function_mapping.items():
                    if func is old:
                        func = new
                        break
        except:
            pass

        s_args = delcapify(e_args)
        result = func(*s_args, **kwargs)

        result = cls(result, **e_args[0].assumptions)

        if (self.expr in (sym.sin, sym.cos, sym.tan, sym.cot, sym.exp,
                          sym.sinh, sym.cosh, sym.tanh, sym.coth)
                and e_args[0].is_phase):
            result.part = ''
        elif self.expr in (sym.atan, sym.atan2):
            result.units = uu.rad
        elif self.expr == sym.diff and isinstance(e_args[1], Expr):
            result.units = e_args[0].units / e_args[1].units
        elif self.expr == Degrees1:
            result.units = uu.deg
        elif self.expr == Radians1:
            result.units = uu.rad
        elif self.expr == sym.integrate:
            if isinstance(e_args[1], Expr):
                result.units = e_args[0].units * e_args[1].units
            elif isinstance(e_args[1], tuple) and isinstance(e_args[1][0], Expr):
                result.units = e_args[0].units * e_args[1][0].units
        elif self.expr is sym.DiracDelta and isinstance(e_args[0], Expr):
            result.units = 1 / e_args[0].units

        return result

    def pdb(self):
        import pdb
        pdb.set_trace()
        return self


class Log10(Function):

    # TODO, figure out how to print as log10(x) rather than
    # the expansion log(x) / log(10).  This will require
    # deferment of the expansion.

    def __call__(self, arg):
        return super(Log10, self).__call__(arg, 10)


def function_wrap(symfunc):
    """Create an Lcapy function from a SymPy function.

    An Lcapy function works with Lcapy expressions and returns an
    Lcapy expression.  It acquires the SymPy docstring.

    """

    name = symfunc.__name__
    docstring = symfunc.__doc__.replace('sympy', 'lcapy')

    newclass = type(name, (Function, ), {'__doc__': docstring})
    return newclass(symfunc)


conjugate = function_wrap(sym.conjugate)

sqrt = function_wrap(sym.sqrt)

exp = function_wrap(sym.exp)

log = function_wrap(sym.log)

log10 = Log10(sym.log)

sin = function_wrap(sym.sin)

cos = function_wrap(sym.cos)

tan = function_wrap(sym.tan)

cot = function_wrap(sym.cot)

asin = function_wrap(sym.asin)

acos = function_wrap(sym.acos)

atan = function_wrap(sym.atan)

atan2 = function_wrap(sym.atan2)

acot = function_wrap(sym.acot)

sinh = function_wrap(sym.sinh)

cosh = function_wrap(sym.cosh)

tanh = function_wrap(sym.tanh)

coth = function_wrap(sym.coth)

asinh = function_wrap(sym.asinh)

acosh = function_wrap(sym.acosh)

atanh = function_wrap(sym.atanh)

acoth = function_wrap(sym.acoth)

gcd = function_wrap(sym.gcd)

sign = function_wrap(sym.sign)

degrees = function_wrap(Degrees1)

radians = function_wrap(Radians1)

diff = function_wrap(sym.diff)

integrate = function_wrap(sym.integrate)

limit = function_wrap(sym.limit)

u = H = heaviside = Heaviside = function_wrap(sym.Heaviside)

delta = DiracDelta = function_wrap(sym.DiracDelta)

Piecewise = function_wrap(sym.Piecewise)

Derivative = function_wrap(sym.Derivative)

Integral = function_wrap(sym.Integral)

Eq = function_wrap(sym.Eq)

Ne = function_wrap(sym.Ne)

Lt = function_wrap(sym.Lt)

Le = function_wrap(sym.Le)

Gt = function_wrap(sym.Gt)

Ge = function_wrap(sym.Ge)

Add = function_wrap(sym.Add)

Mul = function_wrap(sym.Mul)

MatAdd = function_wrap(sym.MatAdd)

MatMul = function_wrap(sym.MatMul)

Sum = function_wrap(sym.Sum)

Min = function_wrap(sym.Min)

Max = function_wrap(sym.Max)

re = function_wrap(sym.re)

im = function_wrap(sym.im)

MatMul = function_wrap(sym.MatMul)

MatAdd = function_wrap(sym.MatAdd)

ui = unitimpulse = UnitImpulse = function_wrap(UnitImpulse1)

us = unitstep = UnitStep = function_wrap(UnitStep1)

rect = function_wrap(rect1)

dtrect = function_wrap(dtrect1)

dtsign = function_wrap(dtsign1)

sinc = sincn = function_wrap(sincn1)

sincu = function_wrap(sincu1)

psinc = function_wrap(psinc1)

tri = function_wrap(tri1)

trap = function_wrap(trap1)

ramp = function_wrap(ramp1)

rampstep = function_wrap(rampstep1)

besseli = function_wrap(sym.besseli)

besselj = function_wrap(sym.besselj)

function_mapping = {sym.Heaviside: us,
                    sym.DiracDelta: ui,
                    rect1: dtrect,
                    sym.sign: dtsign}

from .expr import Expr, expr, delcapify  # nopep8



"""This module provides the SubNetlist class.  It provides circuit
analysis for a single domain quantity.  It is for internal use only.
The Netlist class is for users.

Copyright 2020--2023 Michael Hayes, UCECE

"""

from .symbols import omega
from .state import state
from .mna import MNA
from .netlistmixin import NetlistMixin
from .netlistsimplifymixin import NetlistSimplifyMixin
from .netfile import NetfileMixin


class SubNetlist(NetlistMixin, NetlistSimplifyMixin, NetfileMixin):
    """This is a representation of a netlist for a particular
    transformation domain, such as ac, dc, transient, or noise.  It is
    for internal use only.  Unlike Netlist, SubNetlist is not mutable.
    """

    def __new__(cls, netlist, kind):

        kinds = ('t', 'dc', 's', 'time', 'ivp', 'laplace')
        if not isinstance(kind, str) or kind[0] == 'n':
            pass
        elif kind not in kinds:
            raise ValueError('Expected one of %s for kind, got %s' %
                             (', '.join(kinds), kind))

        obj = netlist.select(kind=kind)
        # Need own context to avoid conflicts with Vn1 and Vn1(s), etc.
        obj.context = state.new_context()
        obj.kind = kind
        obj.__class__ = cls
        obj._analysis = obj.analyse()
        obj.solver_method = netlist.solver_method
        return obj

    def __init__(self, netlist, kind):
        """ kind can be 't', 'dc', 's', 'time', 'ivp', 'n*' or omega,
        where 'n*' is a noise identifer and omega is an angular frequency."""

        self.mna = MNA(self, self.solver_method)

    def get_I(self, name):
        """Current through component"""

        return self.mna.Idict[name].canonical()

    def get_i(self, name):
        """Time-domain current through component"""

        return self.get_I(name).time()

    def get_Vd(self, Np, Nm=None):
        """Voltage drop between nodes"""

        return (self.mna.Vdict[Np] - self.mna.Vdict[Nm]).canonical()

    def get_vd(self, Np, Nm=None):
        """Time-domain voltage drop between nodes"""

        return self.get_Vd(Np, Nm).time()



"""This module provides the Opts class for handling schematic options.

Copyright 2021--2023 Michael Hayes, UCECE
"""

from .latex import latex_format_label


class Opts(dict):

    def __init__(self, arg=None):

        if arg is None:
            return

        if isinstance(arg, str):
            self.add(arg)
            return

        for key, val in arg.items():
            self[key] = val

    def add(self, string):

        def split(s):
            """Split a string by ',' except if in braces"""
            parts = []
            bracket_level = 0
            current = []
            for c in (s + ','):
                if c == ',' and bracket_level == 0:
                    parts.append(''.join(current))
                    current = []
                else:
                    if c == '{':
                        bracket_level += 1
                    elif c == '}':
                        bracket_level -= 1
                    current.append(c)
            if bracket_level != 0:
                raise ValueError('Mismatched braces for ' + s)
            return parts

        if string == '':
            return

        for part in split(string):
            part = part.strip()
            if part == '':
                continue

            fields = part.split('=')
            key = fields[0].strip()
            arg = '='.join(fields[1:]).strip() if len(fields) > 1 else ''
            if arg in ('true', 'True'):
                arg = True
            elif arg in ('false', 'False'):
                arg = False

            if key == 'def':
                if key not in self:
                    self[key] = []
                self[key].append(arg)
            else:
                self[key] = arg

    def __str__(self):
        return self.format()

    @property
    def size(self):

        return float(self.get('size', 1))

    def format(self):

        def fmt(key, val):
            if val == '':
                return key
            return '%s=%s' % (key, val)

        return ', '.join([fmt(key, val) for key, val in self.items()])

    def copy(self):

        return self.__class__(super(Opts, self).copy())

    def remove(self, arg):
        self.pop(arg)

    def strip(self, *args):

        stripped = Opts()
        for opt in args:
            if opt in self:
                stripped[opt] = self.pop(opt)
        return stripped

    def strip_voltage_labels(self):
        """Strip voltage labels."""

        return self.strip('v', 'vr', 'v_', 'v^', 'v_>', 'v_<', 'v^>', 'v^<')

    def strip_current_labels(self):
        """Strip current and flow labels."""

        return self.strip('i', 'ir', 'i_', 'i^', 'i_>', 'i_<', 'i^>', 'i^<',
                          'i>_', 'i<_', 'i>^', 'i<^',
                          'f', 'fr', 'f_', 'f^', 'f_>', 'f_<', 'f^>', 'f^<',
                          'f>_', 'f<_', 'f>^', 'f<^')

    def strip_labels(self):

        return self.strip('l', 'l^', 'l_')

    def strip_all_labels(self):
        """Strip voltage, current, flow, and labels."""

        stripped = Opts()

        stripped.update(self.strip_voltage_labels())
        stripped.update(self.strip_current_labels())
        stripped.update(self.strip_labels())

        return stripped

    def as_list(self, ignore=None, **kwargs):

        if ignore is None:
            ignore = []

        for key, val in self.items():
            # Override with component opts
            kwargs[key] = val

        def fmt(key, val):
            return '%s=%s' % (key, latex_format_label(val))

        return [fmt(key, val) for key, val in kwargs.items() if key not in ignore and '.' not in key]



"""
Lcapy is a Python library for symbolic linear circuit and signal analysis.

Lcapy can analyse circuits described with netlists using modified nodal
analysis.  See lcapy.netlist

Alternatively, Lcapy can analyse networks and circuits formed by
combining one, two, and three port networks.  See lcapy.oneport

For detailed documentation see http://lcapy.readthedocs.io/en/latest

Copyright 2014--2022 Michael Hayes, UCECE
"""
from __future__ import absolute_import, print_function
from sympy import Symbol
from sympy.core.sympify import converter
# This must be imported early to avoid circular import with expr
from .functions import *
from .units import volts, amperes, ohms, siemens, watts
from .state import state
from .inverse_dft import *
from .dft import *
from .differenceequation import *
from .dltifilter import *
from .differentialequation import *
from .ltifilter import *
from .discretetime import *
from .phasor import phasor, phasor_ratio
from .normfexpr import Fexpr
from .normomegaexpr import Omegaexpr
from .omegaexpr import omegaexpr
from .cexpr import *
from .texpr import *
from .uexpr import *
from .sexpr import *
from .fexpr import *
from .expr import *
from .simulator import *
from .randomnetwork import *
from .nettransform import *
from .laplace import *
from .dtstatespace import *
from .statespace import *
from .vector import *
from .tmatrix import *
from .smatrix import *
from .matrix import *
from .sym import *
from .printing import *
from .susceptance import susceptance
from .reactance import reactance
from .inductance import inductance
from .capacitance import capacitance
from .conductance import conductance
from .resistance import resistance
from .transfer import transfer
from .impedance import impedance
from .admittance import admittance
from .current import current, noisecurrent, phasorcurrent
from .voltage import voltage, noisevoltage, phasorvoltage
from .twoport import *
from .oneport import *
from .circuit import *
from .symbols import *
from .nodalanalysis import *
from .loopanalysis import *
from .exprclasses import *
from .seqclasses import *

import sys
import pkg_resources
del absolute_import, print_function


name = "lcapy"


__version__ = pkg_resources.require('lcapy')[0].version
lcapy_version = __version__

if sys.version_info[0] == 2 and sys.version_info[1] < 6:
    raise ImportError("Python Version 2.6 or above is required for Lcapy.")
else:  # Python 3
    pass
    # Here we can also check for specific Python 3 versions, if needed

del sys

# Do not import units.u since this will conflict with unit step


def show_version():
    """Show versions of Lcapy, SymPy, NumPy, MatplotLib, SciPy, and Python."""

    from sys import version as python_version
    from sympy import __version__ as sympy_version
    from numpy import __version__ as numpy_version
    from scipy import __version__ as scipy_version
    from matplotlib import __version__ as matplotlib_version

    print('Python: %s\nSymPy: %s\nNumPy: %s\nMatplotlib: %s\nSciPy: %s\nLcapy: %s' %
          (python_version, sympy_version, numpy_version,
           matplotlib_version, scipy_version, lcapy_version))


# The following is to help sympify deal with j.
# A better fix might be to define an Lcapy class for j and to
# use the __sympy_ method.
converter['j'] = j
converter[Symbol('j')] = j
del converter, Symbol



"""This module provides conductance support.

Copyright 2021--2022 Michael Hayes, UCECE

"""
from .expr import expr
from .units import u as uu
from warnings import warn


def conductance(arg, **assumptions):
    """Create an admittance class for the specified conductance.

    The conductance, G, is defined as the real part of the admittance:

    Y(omega) = G(omega) + j B(omega)"""

    expr1 = expr(arg, frequency=True, **assumptions)
    if expr1.is_imaginary:
        warn('Conductance %s should be real' % expr1)

    expr1.units = uu.siemens
    return expr1



class ExprDomain(object):

    is_sequence = False

    def _class_get(self, quantity, domain):

        from .exprmap import exprmap
        return exprmap(quantity, domain)

    def _class_by_quantity(self, quantity, domain=None):

        if domain is None:
            domain = self.domain
        return self._class_get(quantity, domain)

    def _class_by_domain(self, domain):

        return self._class_get(self.quantity, domain)

    def as_quantity(self, quantity):

        if quantity == 'voltage':
            return self.as_voltage()
        elif quantity == 'current':
            return self.as_current()
        elif quantity == 'impedance':
            return self.as_impedance()
        elif quantity == 'admittance':
            return self.as_admittance()
        elif quantity == 'transfer':
            return self.as_transfer()
        elif quantity == 'power':
            return self.as_power()
        elif quantity == 'undefined':
            return self.as_expr()
        raise ValueError('Unknown quantity %s for %s' % (quantity, self))

    def as_domain(self, domain):

        if domain == 'time':
            return self.as_time()
        elif domain == 'laplace':
            return self.as_laplace()
        elif domain == 'fourier':
            return self.as_fourier()
        elif domain == 'phasor':
            return self.as_phasor()
        elif domain == 'angular fourier':
            return self.as_angular_fourier()
        elif domain == 'frequency response':
            return self.as_frequency_response()
        elif domain == 'angular frequency response':
            return self.as_angular_frequency_response()
        raise ValueError('Unknown domain %s for %s' % (domain, self))

    def as_voltage(self):
        return self._class_by_quantity('voltage')(self)

    def as_current(self):
        return self._class_by_quantity('current')(self)

    def as_admittance(self):
        return self._class_by_quantity('admittance')(self)

    def as_impedance(self):
        return self._class_by_quantity('impedance')(self)

    def as_transfer(self):
        return self._class_by_quantity('transfer')(self)

    def as_power(self):
        return self._class_by_quantity('power')(self)

    def as_expr(self):
        return self

    def as_constant(self):
        from .cexpr import cexpr

        if not self.is_unchanging:
            raise ValueError('Expression %s is not constant' % self)
        return self._class_by_quantity(self.quantity)(self)(cexpr(self))

    def as_superposition(self):
        from .superpositionvoltage import SuperpositionVoltage
        from .superpositioncurrent import SuperpositionCurrent
        from .super import Superposition

        if self.is_voltage:
            return SuperpositionVoltage(self)
        elif self.is_current:
            return SuperpositionCurrent(self)
        elif self.is_undefined:
            return Superposition(self)
        raise ValueError(
            'Can only convert voltage or current to superposition')

    def change(self, arg, domain=None, units_scale=None, **assumptions):
        """Change expression class."""

        if domain is None:
            domain = self.domain

        if domain == 'constant':
            from .expr import expr

            # Allow changing of constants, e.g., V1 to 5 * t
            domain = expr(arg).domain

        quantity = self.quantity

        cls = self._class_by_quantity(quantity, domain)
        ret = cls(arg, **assumptions)

        if units_scale is not None:
            ret.units = self.units * units_scale
        return ret



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_nodal(self):

        a = Circuit("""
        V 1 0 {v(t)}
        R1 1 2
        L 2 3
        R2 3 0_3
        W 0 0_3
        W 3 3_a
        C 3_a 0_4
        W 0_3 0_4""")

        na = a.nodal_analysis()

        na_eqs = na.nodal_equations()

        self.assertEqual(na_eqs[1].lhs, expr('v1(t)'),
                         'nodal_equations()[1].lhs')
        self.assertEqual(na_eqs[1].rhs, expr('v(t)'),
                         'nodal_equations()[1].rhs')



from lcapy import Circuit
from lcapy.circuitgraph import Edges, CircuitGraph


class LadderNetworkMaker:
    """This class converts a circuit into an unbalanced ladder network."""

    def __init__(self, cct, debug=False):

        self.cct = cct
        self.cg = CircuitGraph.from_circuit(cct)
        self.debug = debug

    def _find(self, N1p, N1m, N2p, N2m):
        """Convert a circuit into an unbalanced ladder network.
        The input port is defined by the nodes `N1p` and `N1m`.
        The output port is defined by the nodes `N2p` and `N2m`.

        The nodes `N1p` and `N1m` must be the same.

        This returns `None` if the network does not have a ladder
        topology.  Otherwise, it returns a list of alternating series
        and shunt paths, where a path is one or more edges.  If there
        is no initial series path, it is assigned `[]`.

        """

        # Rename nodes to the unique nodes used in the circuitgraph
        N1p, N1m, N2p, N2m = self.cg.canonical_nodes(N1p, N1m, N2p, N2m)

        parts = []

        if N1m != N2m:
            return parts

        cg = self.cg
        node = N1p
        initial = True

        while True:

            # Can have multiple series paths to the next node
            # Let's assume only a single path
            opts = []

            edges = cg.node_edges(node)

            if len(edges) != 1:
                if initial:
                    # No initial series branch
                    parts.append([])
                else:
                    return []
            else:
                edge = edges[0]

                path = self.cg.series_path(edge, None, N2p)
                if path == []:
                    return []
                if self.debug:
                    print('series: ' + str(path))
                opts.append(path)
                self.cg.remove_edges(path)
                edge = path[-1]
                node = edge.to_node
                parts.append(opts)

            initial = False

            # _Find all the parallel paths
            opts = []
            edges = cg.node_edges(node)
            for edge in edges:
                path = self.cg.series_path(edge, N2m, N2p)
                if path != []:
                    if self.debug:
                        print('parallel: ' + str(path))
                    opts.append(path)
                    self.cg.remove_edges(path)
            parts.append(opts)

            if node == N2p:
                return parts

    def make(self, N1p, N1m, N2p, N2m):
        """Return two-port unbalanced ladder network or `None` if the netlist
        does not have a ladder topology between the specified nodes.

        The input port is defined by the nodes `N1p` and `N1m`.
        The output port is defined by the nodes `N2p` and `N2m`.

        The nodes `N1p` and `N1m` must be the same."""

        if self.cct.dependent_sources != []:
            return None
        if self.cct.mutual_couplings != []:
            return None
        if self.cct.transformers != []:
            return None

        foo = self._find(N1p, N1m, N2p, N2m)

        if foo == []:
            return None

        if foo[-1] == []:
            foo = foo[0:-1]

        from lcapy.oneport import Ser, Par
        from lcapy.twoport import Ladder, LadderAlt

        args = []
        for p in foo:
            pars = []
            for q in p:
                cpts = [
                    self.cct[q1.cpt_name].cpt for q1 in q if q1.cpt_name[0] != 'W']
                if len(cpts) != 1:
                    cpt = Ser(*cpts)
                else:
                    cpt = cpts[0]
                pars.append(cpt)

            if len(pars) == 0:
                cpt = None
            elif len(pars) == 1:
                cpt = pars[0]
            else:
                cpt = Par(*pars)

            args.append(cpt)

        if args[0] is None:
            return LadderAlt(*args[1:])

        return Ladder(*args)


def test():

    cct = Circuit("""
P1  1   0   ; down
R_1 1   2   ; right
C_1 2   3   ; right
C_2 3   4   ; down
L_1 4   0_2 ; down
C_3 3   5   ; right
C_4 5   6   ; down
L_2 6   0_3 ; down
C_5 5   7   ; right
C_7 7   8   ; down
L_3 8   0_4 ; down
C_6 7   9   ; right
R_2 9   0_5 ; down
P2  9_a 0_6 ; down
W   0   0_2 ; right
W   0_2 0_3 ; right
W   0_3 0_4 ; right
W   0_4 0_5 ; right
W   0_5 0_6 ; right
W   9   9_a ; right
""")

    lm = LadderNetworkMaker(cct)
    tp = lm.make(1, 0, 9, 0)
    print(tp)
    return tp



import pdb
from struct import unpack
from glob import glob
from os.path import basename
import imghdr


def get_image_size(filename):

    with open(filename, 'rb') as fhandle:
        header = fhandle.read(24)
        if len(header) != 24:
            raise ValueError('Truncated file')
        if imghdr.what(filename) == 'png':
            check = unpack('>i', header[4:8])[0]
            if check != 0x0d0a1a0a:
                raise ValueError('Not PNG file')
            width, height = unpack('>ii', header[16:24])
    return width, height


def show_image(sch_filename):

    png_filename = sch_filename.replace('.sch', '.png')
    try:
        width, height = get_image_size(png_filename)
    except:
        return

    width_cm = width / 120

    label = basename(sch_filename)

    url = r'https://github.com/mph-/lcapy/tree/master/doc/' + sch_filename

    link = "`%s <%s>`_" % (label, url)

    print(
        """
%s

.. literalinclude:: %s

.. image:: %s
   :width: %.1fcm
        """ % (link, sch_filename, png_filename, width_cm))


class Category:

    def __init__(self, name, basenames):

        self.name = name
        self.basenames = basenames

    def show(self, filename_dict):

        header = self.name
        underline = '=' * len(self.name)
        print(header)
        print(underline + '\n')

        for name in self.basenames:
            filename = filename_dict[name + '.sch']
            show_image(filename)


categories = []

categories.append(Category('Circuits', (
    'E1', 'F1', 'G1', 'H1', 'K2', 'VRL2',
    'VRmesh1',
    'cmos-esd-damage', 'cmos-input-model1', 'cmos-led1', 'cmos-open-drain',
    'cmos-protection1', 'cmos-totem2', 'cmos_C_load_thevenin',
    'cmos_R_series_C_load_thevenin', 'common-base',
    'ferrite-choke1', 'filter1', 'filter2', 'fit3',
    'non-ideal-transformer-primary', 'non-ideal-transformer')))

categories.append(Category('Block diagrams', ('LTFT', 'domains',
                  'fir5', 'negative-feedback2', 'transforms')))

categories.append(Category('Two port networks', (
    'TPA', 'TPB', 'TPG', 'TPH', 'TPY', 'TPZ', 'series', 'shunt',
    'pisection', 'tsection')))

categories.append(Category('Components', (
    'ADC1', 'arrows', 'antennas', 'batteries', 'buffers',
    'cable-coax', 'cable-tp', 'cable-twinax', 'cable-utp', 'capacitors',
    'connections1', 'connections2', 'CPE1', 'DAC3', 'diodes', 'diodes2',
    'FB1', 'fdopamps', 'flipflops', 'grounds', 'GY1',
    'hsources', 'massspringdamper1', 'meters1',
    'misc', 'multiplexers', 'opamps',
    'resistors1', 'resistors2', 'resistors3', 'resistors4',
    'resistors5', 'resistors6', 'resistors7', 'switches', 'switches2',
    'transformers', 'transistors', 'transistors2',
    'Uinamp', 'Uisoamp', 'Ufdopamp', 'Uopamp',
    'variable1', 'vsources', 'wirestyles', 'XT1')))

categories.append(Category('Chips', ('chips', 'chips2', 'chips3')))

categories.append(
    Category('Shapes', ('Striangle2', 'Sbox1', 'Scircle2', 'circles')))

categories.append(Category('Labels', (
    'annotate1', 'current_labels1', 'current_labels2',
    'flow_labels1', 'labels1',
    'voltage_labels1', 'Rlabels')))

categories.append(Category('Digital circuits', (
    'PIO1', 'PIO-input2', 'PIO-input3', 'PIO-output2',  'PIO-pullup2', 'mcu-bt',
    'cmos-backdrive1', 'cmos-backdrive2', 'I2C2', 'ic1',
    'stepup')))

categories.append(Category('Transistor circuits',
                  ('nmosfet_simple_amplifier', 'mosfet_push_pull_load',
                   'BH-mag')))

categories.append(Category('Opamp circuits', (
    'opamp-differential-amplifier1',
    'opamp-inverting-amplifier1',
    'opamp-inverting-integrator',
    'opamp-transimpedance-amplifier1',
    'opamp-transimpedance-amplifier2',
    'inamp-amplifier1', 'fdopamp-amplifier1',
    'opamp-displacement-current-sensor-noise-model1',
    'opamp-displacement-current-sensor1',
    'opamp-noninverting-amplifier-noisy', 'opamp-piezo-amplifier1',
    'opamp-transimpedance-amplifier-with-voltage-gain1',
    'pierce-oscillator', 'sallen-key-lpf1', 'multiple-feedback-lpf',
    'guard1', 'shield-ground', 'shield-guard',
    'opamp-displacement-current-sensor-noise-model1',
    'opamp-displacement-current-sensor1',
    'opamp-noninverting-amplifier-noisy')))


categories.append(Category('Transducers',
                           ('KLM2', 'KLM3', 'accelerometer1')))

categories.append(Category('Miscellaneous', ('colors1', )))

filename_dict = {}
filenames = glob('examples/**/*.sch', recursive=True)
filenames.sort(key=lambda x: basename(x))
for filename in filenames:
    filename_dict[basename(filename)] = filename


print("""
.. _schematic gallery:

==================
Schematic gallery
==================

The images in this section are bit-mapped PNG images with DPI=300.  Higher quality images can be generated using PDF or PGF formats.

""")

for category in categories:
    category.show(filename_dict)



"""This module creates sequences.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from .expr import expr
from .sequence import Sequence
from .nexpr import n
from .utils import isiterable


def seq(arg, ni=None, origin=None, domain=None):
    """Create a Sequence from a tuple, list, ndarray, or str.

    >>> a = seq((1, 2, 3))

    The sequence indices are specified with the optional `ni` argument.
    For example:

    >>> a = seq((1, 2, 3, 4), (-1, 0, 1, 2))

    If the `ni` argument is not specified, the sequence indices
    are enumerated from 0.

    With a string argument, an underscore indicates the zero sequence
    index:

    >>> a = seq('{1, _2, 3, 4}')

    The sequence indices can be found using the `n` attribute.
    This returns a list.

    >>> a = seq('{1, _2, 3, 4}').n
    [-1, 0, 1, 2]

    This function creates a discrete-time domain sequence.  See also
    `nseq` to create a discrete-time domain sequence, `kseq` to create
    a discrete-Fourier domain sequence, and `zseq` to create a
    z-domain sequence.

    """

    if domain is None:
        # For backward-compatibility
        # TODO: If find a z in the arg, then create a zseq.
        return nseq(arg, ni, origin)
    elif domain is n:
        return nseq(arg, ni, origin)
    elif domain is k:
        return kseq(arg, ni, origin)
    elif domain is z:
        return zseq(arg, ni, origin)
    else:
        raise ValueError('Unknown domain %s' % domain)


from .nseq import nseq  # nopep8
from .kseq import kseq  # nopep8
from .zseq import zseq  # nopep8
from .symbols import n, k, z  # nopep8



"""This module provides the ZDomainSequence class to represent
z-domain sequences.

Copyright 2021 Michael Hayes, UCECE

"""

from .domains import ZDomain
from .sequence import Sequence
from .sym import zsym

__all__ = ('zseq', )


class ZDomainSequence(ZDomain, Sequence):
    """z-domain sequence."""

    var = zsym
    domain = 'Z'

    def IZT(self):
        """Calculate inverse z-transform and return as sequence."""

        from .zexpr import z

        results = []
        vals = self.vals
        N = len(vals)
        for ni in range(N):
            result = vals[ni] * z**ni
            result = result.change(result, domain='discrete time')
            results.append(vals[ni] * z**ni)

        return self.change(results, domain='discrete time')


def zseq(arg, ni=None, origin=None):
    """Create a discrete-time domain Sequence from a tuple, list, ndarray, or str.

    >>> a = zseq((1, 2, 3))

    The sequence indices are specified with the optional `ni` argument.
    For example:

    >>> a = zseq((1, 2, 3, 4), (-1, 0, 1, 2))

    If the `ni` argument is not specified, the sequence indices
    are enumerated from 0.

    With a string argument, an underscore indicates the zero sequence
    index:

    >>> a = zseq('{1, _2, 3, 4}')

    The sequence indices can be found using the `n` attribute.
    This returns a list.

    >>> a = zseq('{1, _2, 3, 4}').n
    [-1, 0, 1, 2]
    """

    return ZDomainSequence(arg, ni, origin)



"""This module provides the ConstantDomainExpression class to
represent constant expressions.

Note there are two types:
1. ConstantTimeDomainExpression for signals
2. ConstantFrequencyResponseDomainExpression for transfer functions and immitances.

Note, impedance(3)(s) gives 3 but voltage(3)(s) gives 3 / s.
Similarly, voltage(3)(t) gives 3 but impedance(3)(t) gives 3 * delta)(t)

Copyright 2014--2023 Michael Hayes, UCECE

"""

from .expr import Expr, expr_make
from .domains import ConstantDomain, ConstantTimeDomain, ConstantFrequencyResponseDomain
from .sym import symbols_find
from .voltagemixin import VoltageMixin
from .currentmixin import CurrentMixin
from .admittancemixin import AdmittanceMixin
from .impedancemixin import ImpedanceMixin
from .transfermixin import TransferMixin

__all__ = ('ConstantDomainExpression', 'ConstantTimeDomainExpression',
           'ConstantFrequencyResponseDomainExpression', 'cexpr')


class ConstantExpr(Expr):

    def __init__(self, val, **assumptions):

        symbols = symbols_find(val)
        for symbol in ('s', 'omega', 't', 'f'):
            if symbol in symbols:
                raise ValueError(
                    'Constant expression %s cannot depend on %s' % (val, symbol))

        assumptions['dc'] = True
        super(ConstantExpr, self).__init__(val, **assumptions)

    def as_expr(self):
        return ConstantDomainExpression(self)

    def rms(self):
        """Return root mean square."""
        return self

    def fourier(self):
        """Convert to Fourier domain representation."""

        return self.time().fourier()

    def angular_fourier(self):
        """Convert to angular Fourier domain representation."""

        return self.time().angular_fourier()

    def frequency_response(self):
        """Convert to  frequency response domain representation."""

        return self.change(self, domain='angular frequency response')

    def angular_frequency_response(self):
        """Convert to angular frequency response domain representation."""

        return self.change(self, domain='angular frequency response')

    def canonical(self, factor_const=True):
        # Minor optimisation
        return self

    def phasor(self, omega=None, **assumptions):

        from .phasor import PhasorDomainExpression

        return PhasorDomainExpression.from_constant(self, omega=omega,
                                                    **assumptions)

    def phasor_ratio(self, omega=None, **assumptions):

        return self.as_laplace(**assumptions).phasor_ratio(omega=omega,
                                                           **assumptions)

    def time(self, **assumptions):
        """Convert to time domain representation."""

        return self.change(self, domain='time', **assumptions)

    def laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        return self.time().laplace(**assumptions)
        # return self.change(self, domain='laplace', **assumptions)

    def as_laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        return self.change(self, domain='laplace', **assumptions)

    def response(self, xvector, tvector):
        """Evaluate response to input signal `xvector` at times
        `tvector`.  This returns a NumPy array."""

        from .sexpr import s
        # This can be optimized!
        return self(s).response(xvector, tvector)

    def discretize(self, method=None, alpha=0.5):

        return self


class ConstantDomainExpression(ConstantDomain, ConstantExpr):
    """Constant real expression or symbol.

    If symbols in the expression are known to be negative, use
    ConstantDomainExpression(expr, positive=False)

    """
    pass


class ConstantTimeDomainExpression(ConstantTimeDomain, ConstantExpr):
    """This represents constant voltage, current, voltage-squared, and
    current-squared signals.

    Note, impedance(3)(s) gives 3 but voltage(3)(s) gives 3 / s.

    """

    def _class_by_quantity(self, quantity, domain=None):

        if quantity == 'undefined' and domain is None:
            return ConstantTimeDomainExpression

        return super(ConstantTimeDomainExpression, self)._class_by_quantity(quantity, domain)

    def phasor_ratio(self, omega=None, **assumptions):

        raise ValueError(
            'Cannot convert %s(%s) to phasor_ratio' %
            (self.__class__.__name__, self))


class ConstantFrequencyResponseDomainExpression(ConstantFrequencyResponseDomain,
                                                ConstantExpr):
    """This represents constant impedance, admittance, and transfer
    functions.

    Note, impedance(3)(s) gives 3 but voltage(3)(s) gives 3 / s.
    """

    def _class_by_quantity(self, quantity, domain=None):

        if quantity == 'undefined' and domain is None:
            return ConstantFrequencyResponseDomainExpression

        return super(ConstantFrequencyResponseDomainExpression, self)._class_by_quantity(quantity, domain)

    def phasor(self, omega=None, **assumptions):

        raise ValueError(
            'Cannot convert %s(%s) to phasor' %
            (self.__class__.__name__, self))

    def laplace(self, **assumptions):
        return self.change(self, domain='laplace', **assumptions)

    def time(self, **assumptions):
        """Convert to time domain."""

        return self.laplace(**assumptions).time(**assumptions)


def cexpr(arg, frequency=False, **assumptions):
    """Create Lcapy constant expression from `arg`.

    By default, `arg` is assumed to be positive.  If symbols in the
    `arg` are known to be negative, use `cexpr(arg, positive=False)`.

    """
    try:
        quantity = arg.quantity
    except:
        quantity = 'undefined'

    if quantity == 'undefined':

        if frequency:
            return ConstantFrequencyResponseDomainExpression(arg, **assumptions)

        # Sit on the fence rather than choosing ConstantTimeDomainExpression
        return ConstantDomainExpression(arg, **assumptions)

    return expr_make('constant', arg, **assumptions)



"""This module defines the StateSpace base class for representing a linear
time-invariant system as a state-space model.

Copyright 2019--2022 Michael Hayes, UCECE

"""

from .matrix import Matrix
from .vector import Vector
from .tmatrix import TimeDomainMatrix
from .expr import expr
from .cache import cached_property
import sympy as sym


class StateSpaceBase(object):

    def __init__(self, A, B, C, D, u=None, y=None, x=None, x0=None):
        """Create time-invariant state-space object where:

        A is Nx x Nx state matrix
        B is Nx x Nu input matrix
        C is Ny x Nx output matrix
        D is Ny x Nu feedthrough matrix

        u is Nu x 1 input vector
        x is Nx x 1 state vector
        x0 is Nx x 1 state initial vector
        y is Ny x 1 state output vector
        """

        A = Matrix(A)
        B = Matrix(B)
        C = Matrix(C)
        D = Matrix(D)

        # Number of state variables (the system order).
        Nx = A.shape[0]

        # Number of inputs
        Nu = B.shape[1]

        # Number of outputs
        Ny = C.shape[0]

        if A.shape[0] != A.shape[1]:
            raise ValueError('A matrix not square')
        if B.shape[0] != Nx:
            raise ValueError('B matrix has wrong dimension')
        if C.shape[1] != Nx:
            raise ValueError('C matrix has wrong dimension')
        if (D.shape[0] != Ny) or (D.shape[1] != Nu):
            raise ValueError('D matrix has wrong dimension')

        if x is not None:
            if x.shape[0] != Nx:
                raise ValueError('x vector has wrong dimension')
        if x0 is not None:
            if x0.shape[0] != Nx:
                raise ValueError('x0 vector has wrong dimension')
        if u is not None:
            if u.shape[0] != Nu:
                raise ValueError('u vector has wrong dimension')
        if y is not None:
            if y.shape[0] != Ny:
                raise ValueError('y vector has wrong dimension')

        self._A = A
        self._B = B
        self._C = C
        self._D = D

        self._u = u
        self._x = x
        self._x0 = x0
        self._y = y

    @classmethod
    def from_ba(cls, b, a, form='CCF'):
        """Create state-space representation from transfer function
        specified with numerator and denominator coefficients.

         Note, state-space representations are not unique and are
        determined by the `form` argument.  Currently this can be
        'CCF' for the controllable canonical form.

        """
        return cls.from_transfer_function_coeffs(cls, b, a, form)

    @classmethod
    def from_transfer_function_coeffs(cls, b, a, form='CCF'):
        """Create state-space representation from transfer function
        specified with numerator and denominator coefficients.

         Note, state-space representations are not unique and are
        determined by the `form` argument.  Currently this can be
        'CCF' for the controllable canonical form.

        """

        if form == 'CCF':
            return cls.from_ba_CCF(b, a)
        elif form == 'OCF':
            return cls.from_ba_OCF(b, a)
        elif form == 'DCF':
            return cls.from_ba_DCF(b, a)
        else:
            raise ValueError(
                'Only CCF, DCF, and OCF forms are currently supported')
        # TODO, add Jordan and real Jordan forms.

    @classmethod
    def from_ba_CCF(cls, b, a):

        b = list(b)
        a = list(a)
        Nb = len(b)
        Na = len(a)

        a0 = a[0]
        if a0 != 1:
            a = [ax / a0 for ax in a]
            b = [bx / a0 for bx in b]

        if Na > Nb:
            b = [0] * (Na - Nb) + b
        if Nb > Na:
            # Need extended state-space representation...
            raise ValueError(
                'Improper transfer function; require derivatives of input')

        Nx = len(a) - 1
        Nu = 1
        Ny = 1

        A = Matrix.zeros(Nx, Nx)
        B = Matrix.zeros(Nx, 1)
        C = Matrix.zeros(1, Nx)
        D = Matrix.zeros(1, 1)

        D[0, 0] = b[0]
        for n in range(Nx):
            C[0, n] = b[Nx - n] - a[Nx - n] * b[0]
        B[-1, 0] = sym.S.One

        for n in range(Nx - 1):
            A[n, n + 1] = sym.S.One
        for n in range(Nx):
            A[-1, n] = -a[Nx - n]
        return cls(A, B, C, D)

    @classmethod
    def from_ba_OCF(cls, b, a):

        # Aobs = Acon.T
        # Bobs = Ccon.T
        # Cobs = Bcon.T
        # Dobs = Dcon

        b = list(b)
        a = list(a)
        Nb = len(b)
        Na = len(a)

        a0 = a[0]
        if a0 != 1:
            a = [ax / a0 for ax in a]
            b = [bx / a0 for bx in b]

        if Na > Nb:
            b = [0] * (Na - Nb) + b
        if Nb > Na:
            # Need extended state-space representation...
            raise ValueError(
                'Improper transfer function; require derivatives of input')

        Nx = len(a) - 1
        Nu = 1
        Ny = 1

        A = Matrix.zeros(Nx, Nx)
        B = Matrix.zeros(Nx, 1)
        C = Matrix.zeros(1, Nx)
        D = Matrix.zeros(1, 1)

        D[0, 0] = b[0]
        for n in range(Nx):
            B[n, 0] = b[n + 1] - a[n + 1] * b[0]
        C[0, 0] = sym.S.One

        for n in range(Nx - 1):
            A[n, n + 1] = sym.S.One
        for n in range(Nx):
            A[n, 0] = -a[n + 1]
        return cls(A, B, C, D)

    @classmethod
    def from_ba_DCF(cls, b, a):

        from .sexpr import tf

        b = list(b)
        a = list(a)
        Nb = len(b)
        Na = len(a)

        if Nb > Na:
            # Need extended state-space representation...
            raise ValueError(
                'Improper transfer function; require derivatives of input')

        H = tf(b, a)

        poles = H._ratfun.poles()
        for p in poles:
            if p.n != 1:
                raise ValueError('Require unique poles')

        Nx = len(a) - 1
        Nu = 1
        Ny = 1

        A = Matrix.zeros(Nx, Nx)
        B = Matrix.ones(Nx, 1)
        C = Matrix.zeros(1, Nx)
        D = Matrix.zeros(1, 1)

        # FIXME
        if Na == Nb:
            D[0, 0] = b[0]
        else:
            D[0, 0] = 0

        for n, p in enumerate(poles):
            A[n, n] = p.expr
            C[n] = H._ratfun.residue(p.expr, poles)

        return cls(A, B, C, D)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    @property
    def A(self):
        """State matrix."""
        return self._A

    @property
    def state_matrix(self):
        """State matrix."""
        return self._A

    @property
    def B(self):
        """Input matrix."""
        return self._B

    @property
    def input_matrix(self):
        """Input matrix."""
        return self._B

    @property
    def C(self):
        """Output matrix."""
        return self._C

    @property
    def output_matrix(self):
        """Output matrix."""
        return self._C

    @property
    def D(self):
        """Feed-through matrix."""
        return self._D

    @property
    def feedthrough_matrix(self):
        """Feed-through matrix."""
        return self._D

    @property
    def state_transition_matrix(self):
        """State transition matrix."""
        return self.phi

    @property
    def transfer_functions(self):
        """System transfer functions.  See also `transfer_function`"""
        return self.G

    @property
    def transfer_function(self):
        """System transfer function for a SISO system.  See also
        `transfer_functions`"""
        return self.G[0]

    @cached_property
    def eigenvalues_dict(self):
        """Dictionary of eigenvalues, the roots of the characteristic
        polynomial (equivalent to the poles of Phi(s)).  The
        dictionary values are the multiplicity of the eigenvalues.

        For a list of eigenvalues use eigenvalues."""

        # Equivalent to _A.sympy.eigenvals()

        return self.characteristic_polynomial().roots()

    @cached_property
    def eigenvalues(self):
        """List of eigenvalues, the roots of the characteristic polynomial
        (equivalent to the poles of Phi(s))."""

        roots = self.eigenvalues_dict
        e = []

        # Replicate duplicated eigenvalues and return as a list.
        for v, n in roots.items():
            for m in range(n):
                e.append(v)
        return ExprList(e)

    @cached_property
    def eigenvectors(self):
        """List of tuples (eigenvalue, multiplicity of eigenvalue,
        basis of the eigenspace) of A."""

        return self._A.eigenvects()

    @cached_property
    def singular_values(self):
        """Vector of singular_values."""

        return Vector(self._A.sympy.singular_values(), rational=False)

    @property
    def Nx(self):
        """Number of state variables (the system order)."""
        return self._A.shape[0]

    @property
    def Nu(self):
        """Number of inputs."""
        return self._B.shape[1]

    @property
    def Ny(self):
        """Number of outputs."""
        return self._C.shape[0]

    @cached_property
    def is_symbolic(self):

        return ((self.A.symbols != {}) or (self.B.symbols != {}) or
                (self.C.symbols != {}) or (self.D.symbols != {}))

    @cached_property
    def is_stable(self):
        """True if system is stable."""

        if self.is_symbolic:
            return None

        for e in self.eigenvalues:
            if e.real > 0:
                return False
        return True

    def controllability_matrix_steps(self, steps=None):
        """Return controllability matrix for specified number of time
        steps."""

        if steps is None:
            steps = self.Nx
        if steps < 0:
            raise ValueError('steps must be positive')

        B = self.B
        A = self.A

        R = B
        Q = B
        for n in range(steps - 1):
            Q = A * Q
            R = R.hstack(R, Q)
        return R

    @cached_property
    def controllability_matrix(self):
        """Return controllability matrix."""

        return self.controllability_matrix_steps()

    @property
    def R(self):
        """Return controllability matrix."""

        return self.controllability_matrix

    @cached_property
    def is_controllable(self):
        """True if system is controllable."""

        R = self.controllability_matrix
        return R.rank() == R.shape[0]

    @cached_property
    def observability_matrix(self):
        """Return observability matrix."""

        C = self.C
        A = self.A

        O = C
        Q = C
        for n in range(self.Nx - 1):
            Q *= A
            O = O.vstack(O, Q)
        return O

    @property
    def O(self):
        """Return observability matrix."""

        return self.observability_matrix

    @cached_property
    def is_observable(self):
        """True if system is observable."""

        O = self.observability_matrix
        return O.rank() == O.shape[1]

    @cached_property
    def reachability_range(self):
        """Return reachability range (image, column space)
        as a list of vectors that span the columnspace
        of the controllability matrix."""

        return self.controllability_matrix.columnspace()

    @cached_property
    def hankel_singular_values(self):

        from numpy import sqrt
        from numpy.linalg import eig

        if not self.is_stable:
            raise ValueError('System not stable')

        Wc = self.controllability_gramian.evaluate()
        Wo = self.observability_gramian.evaluate()

        e, UT = eig(Wc @ Wo)

        # CHECKME for abs
        h = sqrt(abs(e))
        return expr(h, rational=False)

    @cached_property
    def balanced_transformation(self):
        """Return the transformation matrix `T` required to balance the
        controllability and observability gramians.

        `Wob = Tinv.T * Wo * Tinv
        Wcb = T * Wc * T.T`

        where `Tinv = T.inv()`

        """

        from scipy import linalg
        from numpy import sqrt, diag

        if not self.is_stable:
            raise ValueError('System not stable')

        Wc = self.controllability_gramian.evaluate()
        Wo = self.observability_gramian.evaluate()

        # Wc = R.T @ R
        R = linalg.cholesky(Wc)

        Y = R @ Wo @ R.T
        # Y = U @ diag(e) @ U.T
        e, U = linalg.eig(Y, left=True, right=False)

        # e is a vector of squared Hankel singular values
        Einv = diag(e ** -0.25)

        Tinv = R.T @ U @ Einv
        return Matrix(Tinv).inv()

    def balance(self):
        """Return new StateSpace object that has the controllability and
        observability gramians equal to the diagonal matrix with the
        Hankel singular values on the diagonal."""

        T = self.balanced_transformation
        return self.transform(T)

    def transform(self, T):

        Tinv = T.inv()
        Ap = T * self.A * Tinv
        Bp = T * self.B
        Cp = self.C * Tinv

        return self.__class__(Ap, Bp, Cp, self.D,
                              self._u, self._y, self._x, self._x0)

    def reduce(self, elim_states, method='truncate'):
        """Perform model reduction given array `elim_states` of
        states to remove.

        `method` can be `'truncate'` or `'matchdc'`.
        """

        from numpy import arange, array, bool_, hstack, linalg

        if isinstance(elim_states[0], (bool, bool_)):
            elim_states = arange(self.Nx)[elim_states]

        # SymPy fancy indexing requires lists not ndarrays.
        melim = sorted(list(elim_states))
        mkeep = []

        for i in range(0, self.Nx):
            if i not in melim:
                mkeep.append(i)

        # A1 is a matrix of all columns of A to keep
        A1 = self.A[:, mkeep[0]]
        for i in mkeep[1:]:
            A1 = hstack((A1, self.A[:, i]))
        A11 = A1[mkeep, :]
        A21 = A1[melim, :]

        # A2 is a matrix of all columns of A to eliminate
        A2 = self.A[:, melim[0]]
        for i in melim[1:]:
            A2 = hstack((A2, self.A[:, i]))
        A12 = A2[mkeep, :]
        A22 = A2[melim, :]

        C1 = self.C[:, mkeep]
        C2 = self.C[:, melim]
        B1 = self.B[mkeep, :]
        B2 = self.B[melim, :]
        D = self.D

        if method == 'truncate':
            Ar = A11
            Br = B1
            Cr = C1
            Dr = D
        elif method == 'matchdc':
            A22I = linalg.inv(A22)

            Ar = A11 - A12 * A22I * A21
            Br = B1 - A12 * A22I * B2
            Cr = C1 - C2 * A22I * A21
            Dr = D - C2 * A22I * B2
        else:
            raise ValueError(
                "Reduction method %s is not supported.  Try 'matchdc' or 'truncate'" % method)

        x = self._x
        x0 = self._x0
        u = self._u
        y = self._y
        if x is not None:
            x = array(x)[mkeep]
        if x0 is not None:
            x0 = array(x0)[mkeep]
        if u is not None:
            u = array(u)[mkeep]
        if y is not None:
            y = array(y)[mkeep]

        return self.__class__(Ar, Br, Cr, Dr, u, y, x, x0)

    def balance_reduce(self, threshold, method='truncate'):
        """Perform balanced model reduction where the states with hankel
        singular values smaller than `threshold` are removed."""

        hsv = self.hankel_singular_values

        elim_states = hsv.numpy.squeeze() < threshold
        return self.reduce(elim_states, method)

    def subs(self, *args, **kwargs):
        """Return new state space object after substituting variables in A, B,
        C, D matrices, see sympy.subs for usage.

        """

        Anew = self.A.subs(*args, **kwargs)
        Bnew = self.B.subs(*args, **kwargs)
        Cnew = self.C.subs(*args, **kwargs)
        Dnew = self.D.subs(*args, **kwargs)

        return self.__class__(Anew, Bnew, Cnew, Dnew,
                              self._u, self._y, self._x, self._x0)


from .symbols import t, s  # nopep8
from .expr import ExprList  # nopep8



"""This module provides resistance support.

Copyright 2021--2022 Michael Hayes, UCECE

"""
from .expr import expr
from .units import u as uu
from warnings import warn


def resistance(arg, **assumptions):
    """Create an impedance class for the specified resistance.

    The resistance, R, is defined as the real part of the impedance:

    Z(omega) = R(omega) + j * X(omega)"""

    expr1 = expr(arg, frequency=True, **assumptions)
    if expr1.is_imaginary:
        warn('Resistance %s should be real' % expr1)

    expr1.units = uu.ohms
    return expr1



"""
This module converts values into engineering format.

Copyright 2021 Michael Hayes, UCECE
"""
from math import floor, log10


class EngFormatter(object):

    def __init__(self, trim=True, hundreds=False, num_digits=3):
        """If `hundreds` True format like 100 pF rather than 0.1 nF"""

        self.trim = trim
        self.hundreds = hundreds
        self.num_digits = num_digits

    def _fmt(self, valstr, unit='', prefix='', space='', mbox_prefix='',
             mbox_suffix=''):

        if unit == '' and prefix == '':
            return valstr

        return valstr + space + mbox_prefix + prefix + unit + mbox_suffix

    def _do(self, value, unit, prefixes, space='', mbox_prefix='',
            mbox_suffix=''):

        value = value
        if value == 0:
            return self._fmt('0', unit, '', space, mbox_prefix, mbox_suffix)

        m = log10(abs(value))

        if m < -1 or m >= 3.0:
            if not self.hundreds:
                m += 1
            n = int(floor(m / 3))
            k = int(floor(m)) - n * 3
        else:
            n = 0
            k = m - 1

        dp = self.num_digits - k

        idx = n + 5
        if idx < 0:
            idx = 0
            return self._fmt('%e' % value, unit, '', space, mbox_prefix, mbox_suffix)
        elif idx >= len(prefixes):
            idx = len(prefixes) - 1
            return self._fmt('%e' % value, unit, '', space, mbox_prefix, mbox_suffix)

        if dp < 0:
            dp = 0
        fmt = '%%.%df' % dp

        n = idx - 5
        value = value * 10**(-3 * n)

        valstr = fmt % value

        if self.trim:
            # Remove trailing zeroes after decimal point
            valstr = valstr.rstrip('0').rstrip('.')

        return self._fmt(valstr, unit, prefixes[idx], space,
                         mbox_prefix, mbox_suffix)

    def latex_math(self, value, unit):
        """Make latex math-mode string."""

        return '$' + self.latex(value, unit) + '$'

    def latex(self, value, unit):
        """Make latex string."""

        return self._do(value, unit,
                        ('f', 'p', 'n', '$\mu$', 'm', '', 'k', 'M', 'G', 'T'),
                        '\,', r'\mbox{', r'}')

    def str(self, value, unit):
        """Make string."""

        return self._do(value, unit,
                        ('f', 'p', 'n', 'u', 'm', '', 'k', 'M', 'G', 'T'),
                        ' ', '', '')



class AdmittanceSquaredMixin(object):

    quantity = 'admittancesquared'
    quantity_label = 'Admittance^2'
    quantity_units = 'S^2'
    is_admittancesquared = True
    is_squared = True
    is_ratio = True



from lcapy import *
from lcapy.phasor import PhasorDomainExpression
import unittest
import sympy as sym
from lcapy.sym import omega0sym


class LcapyOneportTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_R(self):
        """Check R

        """
        a = R(10)
        self.assertEqual(a.Z, impedance(10), "Z incorrect.")
        self.assertEqual(a.Y * 10, admittance(1), "Y incorrect.")
        self.assertEqual2(a.Voc, 0, "Voc incorrect.")
        self.assertEqual2(a.Isc, 0, "Isc incorrect.")
        self.assertEqual2(a.voc, 0, "voc incorrect.")
        self.assertEqual2(a.isc, 0, "isc incorrect.")

    def test_L(self):
        """Check L

        """
        a = L(10, 5)
        self.assertEqual2(a.Z(s), impedance(10 * s), "Zs incorrect.")
        self.assertEqual2(a.Y(s), admittance(1 / (10 * s)), "Ys incorrect.")
        self.assertEqual2(a.Z(jw), impedance(10 * j * omega), "Z incorrect.")
        self.assertEqual2(a.Y(jw), admittance(
            1 / (10 * j * omega)), "Y incorrect.")
        self.assertEqual2(a.Voc, voltage(-10 * 5 + 0 * s), "Voc incorrect.")
        self.assertEqual2(a.Isc, current(-5 / s), "Isc incorrect.")

    def test_C(self):
        """Check C

        """
        a = C(10, 5)

        self.assertEqual2(a.Z(s), impedance(1 / (10 * s)), "Zs incorrect.")
        self.assertEqual2(a.Y(s), admittance(10 * s), "Ys incorrect.")
        self.assertEqual2(a.Z(jw), impedance(
            1 / (10 * j * omega)), "Z incorrect.")
        self.assertEqual2(a.Y(jw), admittance(10 * j * omega), "Y incorrect.")
        self.assertEqual2(a.Voc, voltage(5 / s), "Voc incorrect.")
        self.assertEqual(a.Isc, current(0 * s + 50), "Isc incorrect.")

    def test_R_series_R(self):
        """Check R + R

        """
        a = R(10) + R(5)
        b = a.simplify()
        self.assertEqual2(b.R, impedance(15), "R incorrect.")
        self.assertEqual2(b.Z, impedance(15), "Z incorrect.")
        self.assertEqual(b.Voc, 0, "Voc incorrect.")
        self.assertEqual(b.Isc, 0, "Isc incorrect.")
        self.assertEqual2(type(b), R, "type incorrect.")

    def test_L_series_L(self):
        """Check L + L

        """
        a = L(10) + L(5)
        b = a.simplify()
        self.assertEqual2(b.L, 15, "L incorrect.")
        self.assertEqual2(b.Z(s), impedance(15 * s), "Zs incorrect.")
        self.assertEqual2(b.Z(jw), impedance(15 * j * omega), "Z incorrect.")
        self.assertEqual2(type(b), L, "type incorrect.")

    def test_C_series_C(self):
        """Check C + C

        """
        a = C(10) + C(15)
        b = a.simplify()
        self.assertEqual2(b.C, 6, "C incorrect.")
        self.assertEqual2(b.Z(s), impedance(1 / (6 * s)), "Zs incorrect.")
        self.assertEqual2(b.Z(jw), impedance(
            1 / (6 * j * omega)), "Z incorrect.")
        self.assertEqual2(type(b), C, "type incorrect.")

    def test_Vdc_series_Vdc(self):
        """Check Vdc + Vdc

        """
        a = Vdc(10) + Vdc(5)
        b = a.simplify()
        self.assertEqual2(b.voc, Vdc(15).voc, "voc incorrect.")
        self.assertEqual(b.Voc.dc, voltage(15), "Voc incorrect.")
        self.assertEqual2(type(b), Vdc, "type incorrect.")
        self.assertEqual2(b.voc, voltage(15), "voc incorrect.")

    def test_Vdc_series_R(self):
        """Check Vdc + R

        """
        a = Vdc(10) + R(5)
        self.assertEqual2(a.voc, Vdc(10).voc, "voc incorrect.")
        self.assertEqual(a.Voc.dc, voltage(10), "Voc incorrect.")
        self.assertEqual(a.V.dc, voltage(10), "Voc incorrect.")
        self.assertEqual(a.I, 0, "I incorrect.")
        self.assertEqual(a.v(0), voltage(10), "Voc incorrect.")
        self.assertEqual(a.i(0), 0, "I incorrect.")

        b = a.norton()
        self.assertEqual2(b.voc, Vdc(10).voc, "voc incorrect.")
        self.assertEqual(b.Voc.dc, voltage(10), "Voc incorrect.")
        self.assertEqual(b.V.dc, voltage(10), "Voc incorrect.")
        self.assertEqual(b.I, 0, "I incorrect.")
        self.assertEqual(b.v(0), voltage(10), "Voc incorrect.")
        self.assertEqual(b.i(0), 0, "I incorrect.")

    def test_Vac_series_R(self):
        """Check Vac + R

        """
        a = Vac(10) + R(5)
        self.assertEqual2(a.voc, Vac(10).voc, "voc incorrect.")
        self.assertEqual(a.Voc.ac[omega0sym],
                         phasorvoltage(10, omega=omega0), "Voc incorrect.")
        self.assertEqual(a.V.ac[omega0sym],
                         phasorvoltage(10, omega=omega0), "Voc incorrect.")
        self.assertEqual(a.I, 0, "I incorrect.")
        self.assertEqual(a.v(0), voltage(10), "Voc incorrect.")
        self.assertEqual(a.i(0), 0, "I incorrect.")

        b = a.norton()
        self.assertEqual2(b.voc, Vac(10).voc, "voc incorrect.")
        self.assertEqual(b.Voc.ac[omega0sym],
                         phasorvoltage(10, omega=omega0), "Voc incorrect.")
        self.assertEqual(b.V.ac[omega0sym],
                         phasorvoltage(10, omega=omega0), "Voc incorrect.")
        self.assertEqual(b.I, 0, "I incorrect.")
        self.assertEqual(b.v(0), voltage(10), "Voc incorrect.")
        self.assertEqual(b.i(0), 0, "I incorrect.")

    def test_R_series_L(self):
        """Check R + L

        """
        a = R(10) + L(5)
        self.assertEqual2(a.Z(s), impedance(5 * s + 10), "Zs incorrect.")

    def test_R_parallel_R(self):
        """Check R | R

        """
        a = R(10) | R(15)
        b = a.simplify()
        self.assertEqual2(b.R, impedance(6), "R incorrect.")
        self.assertEqual2(b.Z, impedance(6), "Z incorrect.")
        self.assertEqual2(type(b), R, "type incorrect.")

    def test_L_parallel_L(self):
        """Check L | L

        """
        a = L(10) | L(15)
        b = a.simplify()
        self.assertEqual2(b.L, 6, "L incorrect.")
        self.assertEqual2(b.Z(s), impedance(6 * s), "Zs incorrect.")
        self.assertEqual2(type(b), L, "type incorrect.")

    def test_C_parallel_C(self):
        """Check C | C

        """
        a = C(10) | C(15)
        b = a.simplify()
        self.assertEqual2(b.C, 25, "C incorrect.")
        self.assertEqual2(b.Z(s), impedance(1 / (25 * s)), "Zs incorrect.")
        self.assertEqual2(type(b), C, "type incorrect.")

    def test_I_parallel_I(self):
        """Check I | I

        """
        a = Idc(10) | Idc(5)
        b = a.simplify()
        self.assertEqual2(b.isc, Idc(15).isc, "isc incorrect.")
        self.assertEqual2(b.Isc.dc, current(15), "Isc incorrect.")
        self.assertEqual2(type(b), Idc, "type incorrect.")
        self.assertEqual2(b.isc, current(15), "isc incorrect.")

    def test_load(self):
        """Check load

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.load(R(30))

        self.assertEqual2(type(b), Ser, "type incorrect.")
        self.assertEqual2(b.Z, impedance(7.5), "Shunt loaded R incorrect Z.")

    def test_open_circuit(self):
        """Check open_circuit

        """

        a = Shunt(R(10) + Vdc(5))
        b = a.open_circuit()

        self.assertEqual2(b.Z, impedance(10), "incorrect Z.")
        self.assertEqual2(
            b.Voc.laplace(), LaplaceDomainVoltage(5 / s), "incorrect V.")

    def test_short_circuit(self):
        """Check short_circuit

        """

        a = Series(R(10) + Vdc(5))
        b = a.short_circuit()

        self.assertEqual2(b.Z, impedance(10), "incorrect Z.")
        self.assertEqual2(
            b.Voc.laplace(), LaplaceDomainVoltage(5 / s), "incorrect V.")

    def test_v(self):
        """Check inverse Laplace for voltage sources"""

        a = Vdc(10)
        self.assertEqual(a.voc, voltage(10), "DC incorrect.")

        a = Vstep(10)
        self.assertEqual2(a.voc, voltage(10 * Heaviside(t)), "Step incorrect.")

        a = Vac(10)
        self.assertEqual2(a.voc, voltage(
            10 * cos(omega0 * t)), "AC incorrect.")

    def test_i(self):
        """Check inverse Laplace for current sources"""

        a = Idc(10)
        self.assertEqual(a.isc, current(10), "DC incorrect.")

        a = Istep(10)
        self.assertEqual2(a.isc, current(10 * Heaviside(t)), "Step incorrect.")

        a = Iac(10)
        self.assertEqual2(a.isc, current(
            10 * cos(omega0 * t)), "AC incorrect.")

    def test_CPE(self):
        """Check CPE

        """
        a = CPE(10, 1)

        self.assertEqual2(a.Z(s), impedance(1 / (10 * s)), "Zs incorrect.")
        self.assertEqual2(a.Y(s), admittance(10 * s), "Ys incorrect.")
        self.assertEqual2(a.Z(jw), impedance(
            1 / (10 * j * omega)), "Zw incorrect.")
        self.assertEqual2(a.Y(jw), admittance(10 * j * omega), "Yw incorrect.")

        a = CPE(10, 0)

        self.assertEqual2(a.Z(s), impedance(1.0 / 10), "Zs incorrect.")
        self.assertEqual2(a.Y(s), admittance(10), "Ys incorrect.")
        self.assertEqual2(a.Z(jw), impedance(1.0 / 10), "Zw incorrect.")
        self.assertEqual2(a.Y(jw), admittance(10), "Yw incorrect.")

        a = CPE(10, 0.5)

        self.assertEqual2(a.Y(s), admittance(10 * sqrt(s)), "Ys incorrect.")

    def simplify1(self):

        a = I(0) + Y(1)
        b = a.simplify()

        self.assertEqual2(b._Z, None, "Z incorrect.")
        self.assertEqual2(b._Y, 1, "Y incorrect.")

    def test_cpt(self):

        self.assertEqual(type(voltage(2).cpt()), Vdc, "-> Vdc")
        #self.assertEqual(type(voltage(cos(t)).cpt()), Vac, "-> Vac")

        self.assertEqual(type(current(2).cpt()), Idc, "-> Idc")
        #self.assertEqual(type(current(cos(t)).cpt()), Iac, "-> Iac")

        self.assertEqual(type(impedance(2).cpt()), R, "-> R")
        self.assertEqual(type(impedance(2 * s).cpt()), L, "-> L")
        self.assertEqual(type(impedance(2 / s).cpt()), C, "-> C")

        self.assertEqual(type(admittance(2).cpt()), G, "-> G")
        self.assertEqual(type(admittance(2 * s).cpt()), C, "-> C")
        self.assertEqual(type(admittance(2 / s).cpt()), L, "-> L")

    def test_C_equation(self):

        a = C('C', 'v0')

        self.assertEqual(a.voltage_equation('I(s)', 's'),
                         voltage('I(s) / (s * C) + v0 / s'),
                         "C voltage_equation in s-domain")
        self.assertEqual(a.current_equation('V(s)', 's'),
                         current('s * C * (V(s) - v0 / s)'),
                         "C current_equation in s-domain")

        self.assertEqual(a.voltage_equation('I(s)', 's'),
                         a.voltage_equation('i(t)', 't')(s),
                         "C current_equation in t-domain")

    def test_L_equation(self):

        a = L('L', 'i0')

        self.assertEqual(a.voltage_equation('I(s)', 's'),
                         voltage('s * L * I(s) - L * i0'),
                         "L voltage_equation in s-domain")
        self.assertEqual(a.current_equation('V(s)', 's'),
                         current('(V(s) + L * i0) / (s * L)'),
                         "L current_equation in s-domain")

        self.assertEqual(a.current_equation('V(s)', 's'),
                         a.current_equation('v(t)', 't')(s),
                         "L current_equation in t-domain")
